# nrewc.tcl --
#
#       Uses the nre package to implement nrewc command which is
#       much like the Unix wc program.
#
# See the file "license.terms" for information on usage and
# redistribution of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#

if {[info commands nrematch] == ""} {
    puts "using nre [package require nre]"
}

proc nrewc {files {ver 1}} {
    set res {}
    set errRes {}
    set totalBytes 0
    set totalWords 0
    set totalLines 0
    foreach fn $files {
	if {[catch {open $fn "r"} f]} {
	    # something went wrong
	    lappend errRes $f
	} else {
	    set tmp [_dowc$ver $fn $f]
	    close $f
	    incr totalLines [lindex $tmp 0]
	    incr totalWords [lindex $tmp 1]
	    incr totalBytes [lindex $tmp 2]
	    lappend res $tmp
	}
    }
    if {[llength $res] > 1} {
	lappend res [list $totalLines $totalWords $totalBytes "total"]
    }
    if {$errRes != ""} {
	puts stderr [join $errRes "\n"]
    }
    return $res
}

# _dowc3 is an implementation the standard version of regsub.
# Note regsub does not work correctly with binary files.
proc _dowc0 {fn f} {
    set size [file size $fn]
    set ftext [read $f $size]
    return [list [regsub -all "\n" $ftext {&} dummy] \
		[regsub -all "\[^ \t\n]+" $ftext {&} dummy] \
		$size $fn]
}

# _dowc1 is an implementation using nrematch to count.
proc _dowc1 {fn f} {
    set size [file size $fn]
    set ftext [read $f $size]
    return [list [nrematch -all "\n" $ftext] \
		[nrematch -all {[^[:space:]]+} $ftext] \
		$size $fn]
}

# _dowc2 is an implementation that uses very little memory by using
# the -tryagain switch.
#
proc _dowc2 {fn f {bufsize 0}} {
    set size [file size $fn]
    set woffset 0
    set loffset 0
    if {$bufsize == 0} {
	set bufsize [fconfigure $f -buffersize]
    }
    set lineCount 0
    set wordCount 0
    set ftext [read $f $bufsize]
    while {![eof $f] && [string length $ftext] != 0} {
	incr lineCount [nrematch -start $loffset -all "\n" $ftext]
	incr wordCount [nrematch -tryagain woffset -start $woffset -all \
			    {[^[:space:]]+} $ftext]
	if {$woffset == -1} {
	    set woffset 0
	    set loffset 0
	    set ftext [read $f $bufsize]
	} else {
	    set loffset [string length $ftext]
	    append ftext [read $f $bufsize]
	}
    }
    incr lineCount [nrematch -start $loffset -all "\n" $ftext]
    incr wordCount [nrematch -start $woffset -all {[^[:space:]]+} $ftext]
    return [list $lineCount $wordCount $size $fn]
}

proc _dowc3 {fn f} {
    return [_dowc2 $fn $f 16]
}

# _dowc4 is an implementation using nrematch -eval and a single pass
# over the data.
proc _dowc4 {fn f} {
    set size [file size $fn]
    set ftext [read $f $size]
    set lineCount 0
    set wordCount [nrematch -all -eval "(?:\[^\[:space:]]+)|(\n)" $ftext \
		       {} {} {} {incr lineCount}]
    # Note that nrematch return the total number of matches.
    # Since we matched once for each word and once for each line
    # wordCount is not set to num of words plus num of lines.
    return [list $lineCount [expr {$wordCount - $lineCount}] $size $fn]
}

# _dowc5 is an implementation using nrematch -all and a single pass
# over the data. Some memory is wasted because a list is built up
# to count the lines.
proc _dowc5 {fn f} {
    set size [file size $fn]
    set ftext [read $f $size]
    set lineVar {}
    set wordCount [nrematch -all -prune "(?:\[^\[:space:]]+)|(\n)" $ftext \
		       {} lineVar]
    set lineCount [llength $lineVar]
    # Note that nrematch return the total number of matches.
    # Since we matched once for each word and once for each line
    # wordCount is not set to num of words plus num of lines.
    return [list $lineCount [expr {$wordCount - $lineCount}] $size $fn]
}

proc wctest {} {
    set files {}
    foreach f [glob *.*] {
	if {![file isdirectory $f]} {
	    lappend files $f
	}
    }
    set res [nrewc $files]

    foreach i {2 3 4 5} {
	set res2 [nrewc $files $i]
	if {[string compare $res $res2] != 0} {
	    puts "version $i in examples/nrewc.tcl had a problem!"
	    foreach s $res s2 $res2 {
		if {[string compare $s $s2] != 0} {
		    puts "normal   is: $s"
		    puts "abnormal is: $s2"
		}
	    }
	}
    }
}
