/*
 * TclNewRegComp and TclNewRegExec
 *
 *	Copyright (c) 1986 by University of Toronto.
 *	Written by Henry Spencer.  Not derived from licensed software.
 *
 *	Permission is granted to anyone to use this software for any
 *	purpose on any computer system, and to redistribute it freely,
 *	subject to the following restrictions:
 *
 *	1. The author is not responsible for the consequences of use of
 *		this software, no matter how awful, even if they arise
 *		from defects in it.
 *
 *	2. The origin of this software must not be misrepresented, either
 *		by explicit claim or by omission.
 *
 *	3. Altered versions must be plainly marked as such, and must not
 *		be misrepresented as being the original software.
 *
 * Beware that some of this code is subtly aware of the way operator
 * precedence is structured in regular expressions.  Serious changes in
 * regular-expression syntax might require a total rethink.
 *
 * *** NOTE: this code has been altered slightly for use in Tcl: ***
 * *** 1. Use ckalloc and ckfree instead of  malloc and free.	 ***
 * *** 2. Add extra argument to regexp to specify the real	 ***
 * ***    start of the string separately from the start of the	 ***
 * ***    current search. This is needed to search for multiple	 ***
 * ***    matches within a string.				 ***
 * *** 3. Names have been changed, e.g. from regcomp to		 ***
 * ***    TclNewRegComp, to avoid clashes with other 		 ***
 * ***    regexp implementations used by applications. 		 ***
 * *** 4. Added errMsg declaration and TclNewRegError procedure	 ***
 * *** 5. Various lint-like things, such as casting arguments	 ***
 * ***	  in procedure calls.					 ***
 *
 * *** NOTE: This code has been altered for use in MT-Sturdy Tcl ***
 * *** 1. All use of static variables has been changed to access ***
 * ***    fields of a structure.                                 ***
 * *** 2. This in addition to changes to TclNewRegError makes the   ***
 * ***    code multi-thread safe.                                ***
 *
 * *** NOTE: This code has been altered for nre by Darrel Scheider ***
 * *** See the nrematch.n man page for a description of changes.

 * SCCS: @(#) regexp.c 1.12 96/04/02 13:54:57
 */

#include <tcl.h>
#include "nre.h"
#include <ctype.h>
#include <string.h>
/* #include "tclPort.h" */

#if 0
/* Define META_ATSIGN if you want '@' to work as an escape character
 * in regular expressions.
 */
#define META_ATSIGN
#endif

/*
 * The variable below is set to NULL before invoking regexp functions
 * and checked after those functions.  If an error occurred then TclNewRegError
 * will set the variable to point to a (static) error message.  This
 * mechanism unfortunately does not support multi-threading, but the
 * procedures TclNewRegError and TclNewGetRegError can be modified to use
 * thread-specific storage for the variable and thereby make the code
 * thread-safe.
 */

static char *errMsg = NULL;

/*
 * The "internal use only" fields in regexp.h are present to pass info from
 * compile to execute that permits the execute phase to run lots faster on
 * simple cases.  They are:
 *
 * regstartnode	 first node that must match or NULL if no good start node.
 * regmust	 string (pointer into program) that match must include, or NULL
 * regmlen	 length of regmust string
 *
 * Regstartnode permits very fast decisions on suitable starting points
 * for a match, cutting down the work a lot.  Regmust permits fast rejection
 * of lines that cannot possibly match.  The regmust tests are costly enough
 * that TclNewRegComp() supplies a regmust only if the r.e. contains something
 * potentially expensive (at present, the only such thing detected is * or +
 * at the start of the r.e., which can involve a lot of backup).  Regmlen is
 * supplied because the test in TclNewRegExec() needs it and TclNewRegComp() is
 * computing it anyway.
 */

/*
 * Structure for regexp "program".  This is essentially a linear encoding
 * of a nondeterministic finite-state machine (aka syntax charts or
 * "railroad normal form" in parsing technology).  Each node is an opcode
 * plus a "next" pointer, possibly plus an operand.  "Next" pointers of
 * all nodes except ORBRANCH implement concatenation; a "next" pointer with
 * a ORBRANCH on both ends of it is connecting two alternatives.  (Here we
 * have one of the subtle syntax dependencies:  an individual ORBRANCH (as
 * opposed to a collection of them) is never concatenated with anything
 * because of operator precedence.)  The operand of some types of node is
 * a literal string; for others, it is a node leading into a sub-FSM.  In
 * particular, the operand of a ORBRANCH node is the first node of the alternative.
 * (NB this is *not* a tree structure:  the tail of the alternative connects
 * to the thing following the set of alternatives.)  The opcodes are:
 */

/* definition	number	opnd?	meaning */
#define	END	0	/* no	End of program. */
#define	BOS	1	/* no	Match "" at beginning of input. */
#define	EOS	2	/* no	Match "" at end of input. */
#define	ANY	3	/* no	Match any one character. */
#define	ANYOF	4	/* byte	Match against character set indexed by byte. */
#define ORBRANCH 5      /* node Used to implement the '|' construct. */
#define	BRANCH	6	/* 1byte+node	Used for the '*', '+', and '?' constructs. */
#define	BACK	7	/* no	Match "", "next" ptr points backward. */
#define	EXACTLY	8	/* str	Match this string (first two bytes are length). */
#define	NOTHING	9	/* no	Match empty string. */
#define	STAR	10	/* 1byte+node	Match this (simple) thing 0 or more times. */
#define	PLUS	11	/* node	Match this (simple) thing 1 or more times. */
#define BIANYOF 12      /* byte Like ANYOF but matches against the builtin
			        character class specified by 'byte'. */
#define ORSTART 13      /* byte Like ANYOF but does not consume a character.
                                Used to start a x|y|... sequence.
                                Does a one character lookahead using 'byte'
                                to indicate a character class just like ANYOF.
                                If the byte's valus is UNOPTIMIZED
                                then each branch must be checked for a match. */
#define REPEAT 14	/* 3bytes+node Used to do a {x,y} on a simple node.
			          The first arg byte is x, the second y. */
#define LREPEAT 15	/* 3bytes+node Used to do a {x,y}? on a simple node.
			          The first arg byte is x, the second y. */
#define	WBOUND	16	/* no	Match word boundary '\b'. */
#define	NWBOUND	17	/* no	Match non-word boundary '\B'. */
#define	SOW	18	/* no	Match start-of-word boundary '\<'. */
#define	EOW	19	/* no	Match end-of-word boundary '\>'. */
#define	BOL	20	/* no	Match "" at beginning of line. */
#define	EOL	21	/* no	Match "" at end of line. */
#define	BOINPUT	22	/* no	Match "" at beginning of input '\G'. */
#define LOOK    23      /* node Mark a lookahead '(?='.
			        Node operand is index of ENDLOOK. */
#define NLOOK   24      /* node Mark a negative lookahead '(?!'.
			        Node operand is index of ENDLOOK. */
#define ENDLOOK 25      /* no   ends a LOOK or NLOOK. */
#define	LBRANCH	26	/* 1byte+node	Lazy form of BRANCH. */
#define	LSTAR	27	/* 1byte+node	Lazy form of STAR. */
#define	LPLUS	28	/* node	Lazy form of PLUS. */
#define	NCEXACTLY 29	/* str	Just like EXACTLY but ignore case. */
#define	OPEN	30	/* byte	Mark this point in input as start of #n
			        where #n is byte operand. */
#define	CLOSE	31	/* byte Analogous to OPEN. */
#define	BACKREF	32	/* byte	Backreference to value captured by
			        corresponding OPEN/CLOSE. */
#define OREND   33      /* none Used to mark the end of a ORSTART */
#define FASTLOOK  34    /* none fast form of LOOK */
#define FASTNLOOK 35    /* none fast form of NLOOK */
#define	ISNOTCHAR 36	/* byte	Match against any character except byte.
			 *      This is just a space optimized form of [^x].
			 */
/*
 * Opcode notes:
 *
 * ORBRANCH   	The set of alternativees constituting a single choice are hooked
 *		together with their "next" pointers, since precedence prevents
 *		anything being concatenated to any individual branch.  The
 *		"next" pointer of the last ORBRANCH in a choice points to the
 *		thing following the whole choice.  This is also where the
 *		final "next" pointer of each individual alternative points; each
 *		alternative starts with the operand node of a ORBRANCH node.
 *
 * BACK		Normal "next" pointers all implicitly point forward; BACK
 *		exists to make loop structures possible.
 *
 * STAR,PLUS	'?', and complex '*' and '+', are implemented as circular
 *		BRANCH structures using BACK.  Simple cases (one character
 *		per match) are implemented with STAR and PLUS for speed
 *		and to minimize recursive plunges.
 *
 * OPEN,CLOSE,BACKREF	...are numbered at compile time.
 */

/*
 * A node is one char of opcode followed by two chars of "next" pointer.
 * "Next" pointers are stored as two 8-bit pieces, high order first.  The
 * value is a positive offset from the opcode of the node containing it.
 * An operand, if any, simply follows the node.  (Note that much of the
 * code generation knows about this implicit relationship.)
 *
 * Using two bytes for the "next" pointer is vast overkill for most things,
 * but allows patterns to get big without disasters.
 */
#define	OP(p)	(*(p))
#if 0
#define	NEXT(p)	(((*((p)+1)&0377)<<8) + (*((p)+2)&0377))
#else
#define	NEXT(p)	((((unsigned char*)p)[1] << 8) + ((unsigned char*)p)[2])
#endif
#define regoptnext(p) (p+NEXT(p))
#define regback(p)    (p-NEXT(p))
#define	OPERAND(p)	((p) + 3)

#define GET_EXACT_LEN(ptr) ((((unsigned char*)ptr)[0] << 8) + ((unsigned char*)ptr)[1])

/*
 * See regmagic.h for one further detail of program structure.
 */

/*
 * Utility definitions.
 */
#define	FAIL(m)	{ TclNewRegError(m); return(NULL); }
#define	MULT	  "*?+{"
#define	ISMULT(c)	(csMult.csset[UCHAR(c)])
#define	META      "*?+^$.[()|\\"
#define	METAWHITE "*?+^$.[()|\\ \n\t\f\v\r"
#define	ISMETA(c)	(csMeta.csset[UCHAR(c)])
#define	ISMETAWHITE(c)	(csMetawhite.csset[UCHAR(c)])

/*
 * Flags to be passed up and down.
 */
#define	HASWIDTH	1 /* Known never to match null string. */
#define	SIMPLE		2 /* Simple enough to be STAR/PLUS/REPEAT operand. */
#define	SPSTART		4 /* Starts with * or +. */
#define	WORST		0 /* Worst case. */

#define CHARMAP_SIZE 256
#if CHARMAP_SIZE == 256
#define MAX_CSTOTAL 32640 /* 255 + 254 + 253 + ... + 1 */
#else
+++ port error: need to calculate MAX_CSTOTAL
#endif


struct re_classset {
  long cscount;
  long cstotal;
  long negated;
  char csset[CHARMAP_SIZE];
};

/*
 * Global work variables for TclNewRegComp().
 */
#define MAX_CLASS_COUNT 64
struct regcomp_state  {
  char *regparse;		/* Input-scan pointer. */
  char *regparseEnd;		/* Points to byte after end of input. */
  char *regcode;		/* Code-emit pointer */
  char *regcodeEnd;
  int extflags;
  int regnpar;		/* () count. */
  int regnclosepar;
  char regparhaswidth[NEWNSUBEXP];
  int regnClassCount;
  struct re_classset classtab[MAX_CLASS_COUNT];
};

/*
 * The first byte of the regexp internal "program" is actually this magic
 * number; the start node begins in the second byte.
 */
#define	MAGIC	0234

/*
 * Forward declarations for TclNewRegComp()'s friends.
 */

static char *		reg _ANSI_ARGS_((int paren, int *flagp,
			    struct regcomp_state *rcstate));
static char *		regatom _ANSI_ARGS_((int *flagp,
					     struct regcomp_state *rcstate,
					     char *chain));
static char *		regbranch _ANSI_ARGS_((int *flagp,
			    struct regcomp_state *rcstate));
static void		regc _ANSI_ARGS_((int b,
			    struct regcomp_state *rcstate));
static char *		reginsert _ANSI_ARGS_((int op,
					       char *opnd,
					       struct regcomp_state *rcstate,
					       int size));
static char *		regnext _ANSI_ARGS_((char *p));
static char *		regnode _ANSI_ARGS_((int op,
			    struct regcomp_state *rcstate));
static void		regrmnode _ANSI_ARGS_((char *node,
					       struct regcomp_state *rcstate,
					       int size));
static void 		regoptail _ANSI_ARGS_((char *p, char *val));
static char *		regpiece _ANSI_ARGS_((int *flagp,
					      struct regcomp_state *rcstate,
					      char *chain));
static void 		regtail _ANSI_ARGS_((char *p, char *val));
static char*            regcpnode _ANSI_ARGS_((char *node,
					      struct regcomp_state *rcstate));
static void             initBuiltinClasses _ANSI_ARGS_((void));
static void eatWhiteSpace _ANSI_ARGS_((struct regcomp_state *rcstate));
static char* parseQuestionMod _ANSI_ARGS_((int *flagp,
					   struct regcomp_state *rcstate,
					   int *parenFlagsPtr,
					   char *chain));
static char* parseMacroMod _ANSI_ARGS_((int *flagp,
					struct regcomp_state *rcstate,
				       int macroOp));
typedef int (re_classtestProc) _ANSI_ARGS_((int));
static void  initBuiltinClass _ANSI_ARGS_((struct re_classset *csp,
					   re_classtestProc *fptr,
					   int negated));
static void initcs _ANSI_ARGS_((struct re_classset *csp, int negated));
static void addclasschar _ANSI_ARGS_((int c,
				      struct regcomp_state *rcstate,
				      struct re_classset *csp));
static int addclassnotchar _ANSI_ARGS_((int c,
					struct re_classset *csp));
static int getPosixRangeClass _ANSI_ARGS_((struct regcomp_state *rcstate,
				           struct re_classset *csp,
					   int* ok));
static void addcs _ANSI_ARGS_((struct re_classset *csp, int c));
static char *addclassnode _ANSI_ARGS_((struct regcomp_state *rcstate,
				       int negated,
				       struct re_classset *csp));
static int calcNextEatenByte _ANSI_ARGS_((struct regcomp_state *rcstate,
					  char *next,
					  struct re_classset *cmap));
static char *findStartNode _ANSI_ARGS_((char *next));
static void optimize _ANSI_ARGS_((struct regcomp_state *rcstate, char* scan));
static char* parseCharRange _ANSI_ARGS_((struct regcomp_state *rcstate,
					 int termChar));
static int getRangeChar _ANSI_ARGS_((struct regcomp_state *rcstate));
static int doSetsIntersect _ANSI_ARGS_((char *csp1, char *csp2));
static char *findStartNode _ANSI_ARGS_((char *next));
static void charMapAdd _ANSI_ARGS_((struct re_classset* dst,
				    CONST struct re_classset* src));
static int calcNextEatenByte _ANSI_ARGS_((struct regcomp_state *rcstate,
					  char *next,
					  struct re_classset *cmap));
static void findCharMap _ANSI_ARGS_((struct regcomp_state *rcstate,
				     char* modptr,
				     struct re_classset *csp));
static void fcdStartOrCheck _ANSI_ARGS_((struct regcomp_state *rcstate,
					 char* orstart));
static void fcdBranchCheck _ANSI_ARGS_((struct regcomp_state *rcstate,
					char* modptr,
					char* b1,
					char* b2));
static int optimizeNodes _ANSI_ARGS_((struct regcomp_state *rcstate,
				      char* scan,
				      char* stopNode));
static int implicitLineAnchorCheck _ANSI_ARGS_((char* scan));
static void optimize _ANSI_ARGS_((struct regcomp_state *rcstate, char* scan));
static char* anyofScan _ANSI_ARGS_((char *s,
				    CONST char* anyofPtr,
				    char* endPtr));
static char* isNotCharScan _ANSI_ARGS_((char *s, int c, char* endPtr));
static char* sowScan _ANSI_ARGS_((char *s, char* endPtr));
static char* wboundScan _ANSI_ARGS_((char *s, char* endPtr, char* startPtr));
static int re_isspace _ANSI_ARGS_((int c));
static int re_isdigit _ANSI_ARGS_((int c));
static int re_isalnum _ANSI_ARGS_((int c));
static int re_isalpha _ANSI_ARGS_((int c));
static int re_isupper _ANSI_ARGS_((int c));
static int re_islower _ANSI_ARGS_((int c));
static int re_isdigit _ANSI_ARGS_((int c));
static int re_isxdigit _ANSI_ARGS_((int c));
static int re_ispunct _ANSI_ARGS_((int c));
static int re_isgraph _ANSI_ARGS_((int c));
static int re_isprint _ANSI_ARGS_((int c));
static int re_iscntrl _ANSI_ARGS_((int c));


#if 0
#define DEBUG
#else
#undef DEBUG
#endif

static char *regprop _ANSI_ARGS_((regexp *r, char *op));

enum {
  CS_SPACE,    /* ispace */
  CS_NOTSPACE,
  CS_DIGIT,    /* isdigit */
  CS_NOTDIGIT,
  CS_WORD,     /* isalnum + '_' */
  CS_NOTWORD,
  CS_UPPER,
  CS_NOTUPPER,
  CS_LOWER,
  CS_NOTLOWER,
  CS_ALPHA,
  CS_NOTALPHA,
  CS_ALPHANUM,
  CS_NOTALPHANUM,
  CS_XDIGIT,
  CS_NOTXDIGIT,
  CS_PUNCT,
  CS_NOTPUNCT,
  CS_GRAPH,
  CS_NOTGRAPH,
  CS_PRINT,
  CS_NOTPRINT,
  CS_CNTRL,
  CS_NOTCNTRL,
  CS_LAST
};

static int builtin_initialized = 0;
static struct re_classset bitab[CS_LAST];
static char* bistr[CS_LAST] = {
#ifdef META_ATSIGN
  "@s",
  "@S",
  "@d",
  "@D",
  "@w",
  "@W",
#else
  "\\s",
  "\\S",
  "\\d",
  "\\D",
  "\\w",
  "\\W",
#endif
  "[[:upper:]]",
  "[^[:upper:]]",
  "[[:lower:]]",
  "[^[:lower:]]",
  "[[:alpha:]]",
  "[^[:alpha:]]",
  "[[:alnum:]]",
  "[^[:alnum:]]",
  "[[:xdigit:]]",
  "[^[:xdigit:]]",
  "[[:punct:]]",
  "[^[:punct:]]",
  "[[:graph:]]",
  "[^[:graph:]]",
  "[[:print:]]",
  "[^[:print:]]",
  "[[:cntrl:]]",
  "[^[:cntrl:]]"
};
static char lcmap[CHARMAP_SIZE];
static struct re_classset csMult; /* used internally '*', '+', '?', '{' */
static struct re_classset csMeta; /* used internally for chars in META */
static struct re_classset csMetawhite; /* used internally for chars in METAWHITE */

/* returns non-zero if 'c' is in the builtin identified by 'idx' */
#define BITEST(idx, c) (bitab[idx].csset[UCHAR(c)])
#define BIANYOFPTR(idx) (&bitab[idx].csset[0])
#define ANYOFPTR(p, idx) ((p) + ((idx) * CHARMAP_SIZE))
#define ANYOFTEST(p, idx, c) (ANYOFPTR(p, idx)[UCHAR(c)])
#define TOLC(c) (lcmap[UCHAR(c)])

/*
 - TclNewRegComp - compile a regular expression into internal code
 *
 * We can't allocate space until we know how big the compiled form will be,
 * but we can't compile it (and thus know how big it is) until we've got a
 * place to put the code.  So we cheat:  we compile it twice, once with code
 * generation turned off and size counting turned on, and once "for real".
 * This also means that we don't allocate space until we are sure that the
 * thing really will compile successfully, and we never have to move the
 * code and thus invalidate pointers into it.  (Note that it has to be in
 * one piece because free() must be able to free it all.)
 *
 * Beware that the optimization-preparation code in here knows about some
 * of the structure of the compiled regexp.
 */
regexp *
TclNewRegComp(exp)
char *exp;
{
  return TclNewRegCompSz(exp, -1, 0);
}

regexp *
TclNewRegCompSz(exp, length, noCase)
char *exp;
int length;
int noCase;
{
	register regexp *r;
	register char *scan;
	register char *longest;
	register int len;
	int flags;
	struct regcomp_state state;
	struct regcomp_state *rcstate= &state;
	char* res;
	size_t regsize;
	size_t classsize;
#define MAX_REGEXP 32768L   /* Probably could be 65535L. */
	char tmpregexp[4096];
	char *compBuff = tmpregexp;

	if (exp == NULL) {
	  FAIL("NULL argument");
	  /* NOTREACHED */
	}

	initBuiltinClasses();
	rcstate->extflags = 0;
	if (noCase) {
	  rcstate->extflags |= REGEXP_NOCASE;
	}
	rcstate->regcodeEnd = tmpregexp+sizeof(tmpregexp);

	if (length == -1) {
	  length = strlen(exp);
	}
	rcstate->regparseEnd = exp + length;

	for (;;) {
	  int oldflags = rcstate->extflags;
	  rcstate->regparse = exp;
	  rcstate->regnpar = 1;
	  rcstate->regnclosepar = 0;
	  rcstate->regnClassCount = 0;
	  rcstate->regcode = compBuff;
	  regc(MAGIC, rcstate);
	  res = reg(0, &flags, rcstate);
	  /* if flags changed need to start over */
	  if (res == NULL) {
	    if (oldflags != rcstate->extflags) {
	      /* found some flags that changed so restart compile */
	      continue;
	    } else if ((rcstate->regcode+16) >= rcstate->regcodeEnd) {
	      /* ran out of room. Try growing compilation buffer
	       * and recompile.
	       */
	      if (compBuff == tmpregexp) {
		compBuff = (char*)ckalloc(MAX_REGEXP);
		rcstate->regcodeEnd = compBuff + MAX_REGEXP;
		errMsg = NULL;
		continue;
	      } else {
		ckfree(compBuff);
		return(NULL);
	      }
	    } else {
	      if (compBuff != tmpregexp) {
		ckfree(compBuff);
	      }
	      return(NULL);
	    }
	  } else {
	    break;
	  }
	}

	optimize(rcstate, compBuff+1);
	
	/* Allocate space. */
	
	regsize = rcstate->regcode - compBuff;
	classsize = (rcstate->regnClassCount * CHARMAP_SIZE) + (sizeof(long)*2);
	r = (regexp *)ckalloc(sizeof(regexp) + regsize + classsize);
	if (r == NULL) {
	  if (compBuff != tmpregexp) {
	    ckfree(compBuff);
	  }
	  FAIL("out of space");
	  /* NOTREACHED */
	}
	r->regexpSize = sizeof(regexp) + regsize + classsize;

	memcpy(r->program, compBuff, regsize);
	if (compBuff != tmpregexp) {
	  ckfree(compBuff);
	  compBuff = NULL;
	}
	if (rcstate->regnClassCount > 0) {
	  unsigned long foo;
	  int i;
	  char* classptr = r->program + (regsize+1);
	  /* make sure it is longword aligned.
	   * This allows for fast intersection checks.
	   */
	  foo = (unsigned long)classptr;
	  foo += sizeof(long) - 1;
	  foo /= sizeof(long);
	  foo *= sizeof(long);
	  classptr = (char*)foo;
	  r->regclasses = classptr;
	  for (i= 0; i < rcstate->regnClassCount; i++) {
	    memcpy(classptr, rcstate->classtab[i].csset, CHARMAP_SIZE);
	    classptr += CHARMAP_SIZE;
	  }
	} else {
	  r->regclasses = NULL;
	}
	

	/* Dig out information for optimizations. */
	r->regmust = NULL;
	r->regmlen = 0;
	r->regflags = rcstate->extflags;
	/* findStartNode is so simple and helpful we do it even if QUICKCOMPILE */
	r->regstartnode = findStartNode(r->program+1);
	r->subexpCount = rcstate->regnpar;
	scan = r->program+1;

	/*
	 * If there's something expensive in the r.e., find the
	 * longest literal string that must appear and make it the
	 * regmust.  Resolve ties in favor of later strings, since
	 * the regstart check works with the beginning of the r.e.
	 * and avoiding duplication strengthens checking.  Not a
	 * strong reason, but sufficient in the absence of others.
	 */
	if (flags&SPSTART && !(rcstate->extflags & REGEXP_QUICKCOMPILE)) {
	  longest = NULL;
	  len = 0;
	  for (; scan != NULL; scan = regnext(scan)) {
	    if (OP(scan) == FASTNLOOK) {
	      /* skip the FASTNLOOK's operand since it is negated */
	      scan = regnext(scan);
	      if (scan == NULL) {
		break;
	      }
	    } else if ((OP(scan) == EXACTLY || OP(scan) == NCEXACTLY) &&
		       scan != r->regstartnode) {
	      unsigned char* lenPtr = (unsigned char*)OPERAND(scan);
	      unsigned int curLen = GET_EXACT_LEN(lenPtr);
	      if ((int)curLen >= len) {
		longest = (char*)(lenPtr + 2);
		len = (int)curLen;
	      }
	    }
	  }
	  if (len > 1) {
	    r->regmust = longest;
	    r->regmlen = len;
	  }
	}
	return(r);
}

#define GROUP_BIT 1
#define CAPTURE_BIT 2
#define RECOMPILE_BIT 4
#define LOOK_BIT 8
#define NLOOK_BIT 16

#define UNOPTIMIZED 255
/*
 - reg - regular expression, i.e. main body or parenthesized thing
 *
 * Caller must absorb opening parenthesis.
 *
 * Combining parenthesis handling with the base level of regular expression
 * is a trifle forced, but the need to tie the tails of the branches to what
 * follows makes it hard to avoid.
 */
static char *
reg(paren, flagp, rcstate)
int paren;			/* Parenthesized? */
int *flagp;
struct regcomp_state *rcstate;
{
	char *ret;
	char *br;
	char *firstbr;
	char *lastbr;
	char *ender;
	int parno = 0;
	int flags;
	int brcount = 0;

	*flagp = HASWIDTH;	/* Tentatively. */

	ret = NULL;
	if (paren) {
	  /* Make an OPEN node, if parenthesized. */
	  if (paren & CAPTURE_BIT) {
	    if (rcstate->regnpar >= NEWNSUBEXP) {
	      FAIL("too many ()");
	      /* NOTREACHED */
	    }
	    parno = rcstate->regnpar;
	    rcstate->regnpar++;
	    rcstate->regparhaswidth[parno] = 0;
	    ret = regnode(OPEN,rcstate);
	    regc(parno, rcstate);
	  } else if (paren & LOOK_BIT) {
	    ret = regnode(LOOK,rcstate);
	  } else if (paren & NLOOK_BIT) {
	    ret = regnode(NLOOK,rcstate);
	  }
	}

	/* Pick up the branches, linking them together. */
	firstbr = regnode(ORSTART,rcstate);
	regc(UNOPTIMIZED, rcstate);
	if (regbranch(&flags,rcstate) == NULL) {
	  return(NULL);
	}
	brcount = 1;
	if (ret != NULL) 
		regtail(ret, firstbr);	/* OPEN -> first. */
	else
		ret = firstbr;
	if (!(flags&HASWIDTH))
		*flagp &= ~HASWIDTH;
	*flagp |= flags&SPSTART;
	lastbr = firstbr;
	while (rcstate->regparse != rcstate->regparseEnd &&
	       *rcstate->regparse == '|') {
		rcstate->regparse++;
		br = regnode(ORBRANCH,rcstate);
		if (regbranch(&flags,rcstate) == NULL) {
		  return(NULL);
		}
		brcount++;
		regtail(lastbr, br);	/* ORBRANCH -> ORBRANCH. */
		if (!(flags&HASWIDTH))
			*flagp &= ~HASWIDTH;
		/* Note we ignore SPSTART from these branches because
		 * we only care about the first one.
		 * The first branch will always be walked.
		 */
		lastbr = br;
	}


	if (brcount == 1) {
	  /* Delete since we only have one choice */
	  regrmnode(lastbr, rcstate, 4);
	} else {
	  /* change last ORBRANCH to a OREND */
	  lastbr[0] = OREND;
	}
	/* Clear the SIMPLE flag */
	*flagp &= ~SIMPLE;
	/* Make a closing node, and hook it on the end. */
	ender = NULL;
	if (paren) {
	  if (paren & CAPTURE_BIT) {
	    rcstate->regnclosepar++;
	    ender = regnode(CLOSE,rcstate);
	    regc(parno, rcstate);
	    if (*flagp & HASWIDTH) {
	      rcstate->regparhaswidth[parno] = 1;
	    }
	  } else if (paren == GROUP_BIT) {
	    /* Just a (?:...) */
	    if (brcount > 1) {
	      ender = regnode(NOTHING,rcstate);
	    }
	  } else {
	    /* clear HASWIDTH since lookahead is always zero-size */
	    *flagp &= ~HASWIDTH;
	    if ((brcount == 1) && (flags&SIMPLE)) {
	      /* Change the LOOK to an optimized form.
	       *.This will tell the matcher to do a quick match
	       * on the simple node that its next points to.
	       */
	      if (ret[0] == LOOK) {
		ret[0] = FASTLOOK;
	      } else {
		ret[0] = FASTNLOOK;
	      }
	    } else {
	      ender = regnode(ENDLOOK,rcstate);
	      ret[1] = 0;
	      ret[2] = 0;
	      regtail(ret, ender);
	    }
	  }
	} else {
	  ender = regnode(END,rcstate);	
	}

	if (ender) {
	  regtail(lastbr, ender);

	  if (brcount > 1) {
	    /* Hook the tails of the branches to the closing node. */
	    /* ORSTART -> ender */
	    regtail(OPERAND(firstbr+1), ender);
	    br = regnext(firstbr);
	    while (br != lastbr) {
	      /* ORBRANCH -> ender */
	      regtail(OPERAND(br), ender);
	      br = regnext(br);
	    }
	    /* OREND -> ender */
	    regtail(OPERAND(lastbr), ender);
	  }
	}

	/* Check for proper termination. */
	if (paren) {
	  if (rcstate->regparse == rcstate->regparseEnd ||
	      *rcstate->regparse++ != ')') {
	    FAIL("unmatched ()");
	    /* NOTREACHED */
	  } else {
	    if (rcstate->extflags & REGEXP_XMODE) {
	      eatWhiteSpace(rcstate);
	    }
	  }
	} else if (rcstate->regparse != rcstate->regparseEnd) {
	  if (*rcstate->regparse == ')') {
	    FAIL("unmatched ()");
	  } else {
	    FAIL("junk on end");	/* "Can't happen". */
	  }
	  /* NOTREACHED */
	}

	return(ret);
}

/*
 - regbranch - one alternative of an | operator
 *
 * Implements the concatenation operator.
 */
static char *
regbranch(flagp, rcstate)
int *flagp;
struct regcomp_state *rcstate;
{
	register char *ret = rcstate->regcode;
	register char *chain;
	register char *latest;
	int flags;

	*flagp = WORST;		/* Tentatively. */

	chain = NULL;
	if (rcstate->extflags & REGEXP_XMODE) {
	  eatWhiteSpace(rcstate);
	}
	while (rcstate->regparse != rcstate->regparseEnd &&
	       *rcstate->regparse != '|' &&
	       *rcstate->regparse != ')') {
		latest = regpiece(&flags, rcstate, chain);
		if (latest == NULL) {
		  return(NULL);
		}
		if (chain == NULL) {	/* First piece. */
		  *flagp |= flags&(HASWIDTH|SPSTART|SIMPLE);
		} else {
		  *flagp |= flags&(HASWIDTH|SPSTART);
		  /* If more than one node than no way this is SIMPLE */
		  *flagp &= ~SIMPLE;
		  regtail(chain, latest);
		}
		chain = latest;
	}
	if (chain == NULL)	/* Loop ran zero times. */
		(void) regnode(NOTHING,rcstate);

	return(ret);
}

/*
 - regpiece - something followed by possible [*+?]
 *
 * Note that the branching code sequences used for ? and the general cases
 * of * and + are somewhat optimized:  they use the same NOTHING node as
 * both the endmarker for their branch list and the body of the last branch.
 * It might seem that this node could be dispensed with entirely, but the
 * endmarker role is not redundant.
 */
static char *
regpiece(flagp, rcstate, chain)
int *flagp;
struct regcomp_state *rcstate;
char* chain;
{
	char *ret;
	char *atomp;
	char op;
	char *next;
	int flags;
	int isLazy = 0;
	int optionalCount = 1;
	int min, max;

	ret = regatom(&flags, rcstate, chain);
	if (ret == NULL) {
	  return(NULL);
	}

	if (rcstate->regparse == rcstate->regparseEnd) {
	  *flagp = flags;
	  return(ret);
	}
	op = *rcstate->regparse;
	if (!ISMULT(op)) {
	  *flagp = flags;
	  return(ret);
	}
	if (!(rcstate->extflags & REGEXP_PERLMODE) && op == '{') {
	  *flagp = flags;
	  return(ret);
	}

	rcstate->regparse++;
	if (op != '{' && rcstate->regparse != rcstate->regparseEnd &&
	    *rcstate->regparse == '?') {
	  isLazy = 1;
	  rcstate->regparse++;
	}

	atomp = ret;
	*flagp = WORST;
	switch (op) {
	case '?' :
	  min = 0;
	  max = 1;
	  break;
	case '*' :
	  min = 0;
	  max = -1;
	  break;
	case '+' :
	  min = 1;
	  max = -1;
	  break;
	  break;
#define MAXRDIGITS 3
#define MAXRVALUE 255
	case '{' : {
	  /* parse {x[,[y]]}[?]. Note that whitepace is not allowed. */
	  int x, y;
	  int count;
	  x = 0;
	  y = -1; /* infinite */

	  count = 0;
	  while (rcstate->regparse != rcstate->regparseEnd &&
		 isdigit(*rcstate->regparse)) {
	    count++;
	    if (count > MAXRDIGITS) {
	      FAIL("repetition count too large");
	      /* NOTREACHED */
	    }
	    x *= 10;
	    x += *rcstate->regparse - '0';
	    rcstate->regparse++;
	  }
	  if (count == 0) {
	    FAIL("missing repetition count");
	    /* NOTREACHED */
	  }
	  if (x > MAXRVALUE) {
	    FAIL("repetition count too large");
	    /* NOTREACHED */
	  }
	  if (rcstate->regparse == rcstate->regparseEnd ||
	      *rcstate->regparse != ',') {
	    y = x;
	  } else {
	    rcstate->regparse++;
	    count = 0;
	    y = 0;
	    while (rcstate->regparse != rcstate->regparseEnd &&
		   isdigit(*rcstate->regparse)) {
	      count++;
	      if (count > MAXRDIGITS) {
		FAIL("repetition count too large");
		/* NOTREACHED */
	      }
	      y *= 10;
	      y += *rcstate->regparse - '0';
	      rcstate->regparse++;
	    }
	    if (count == 0) {
	      y = -1; /* infinite */
	    } else if (y > MAXRVALUE) {
	      FAIL("repetition count too large");
	      /* NOTREACHED */
	    } else if (y < x) {
	      FAIL("second repetition count less than first");
	      /* NOTREACHED */
	    }
	  }
	  if (rcstate->regparse != rcstate->regparseEnd &&
	      *rcstate->regparse == '}') {
	    rcstate->regparse++;
	  } else {
	    FAIL("repetition missing '}'");
	    /* NOTREACHED */
	  }
	  if (rcstate->regparse != rcstate->regparseEnd &&
	      *rcstate->regparse == '?') {
	    isLazy = 1;
	    rcstate->regparse++;
	  }
	  if (x > 0) {
	    /* *flagp |= HASWIDTH; */
	  } else if (y == x) {
	    FAIL("repetition count was 0");
	    /* NOTREACHED */
	  }
	  min = x;
	  max = y;
	  break;
	}
	} /* switch */
	if (max == -1) {
	  *flagp |= SPSTART;
	  if (!(flags&HASWIDTH)) {
	    FAIL("*+ operand could be empty");
	    /* NOTREACHED */
	  }
	}
	if (min > 0 && (flags&HASWIDTH)) {
	  *flagp |= HASWIDTH;
	}
	if (flags&SIMPLE) {
	  if (max == -1 && (min == 0 || min == 1)) {
	    if (min == 0) {
	      if (!reginsert(isLazy?LSTAR:STAR, atomp, rcstate, 4)) {
		return NULL;
	      }
	      OPERAND(atomp)[0] = (char)UNOPTIMIZED;
	    } else {
	      if (!reginsert(isLazy?LPLUS:PLUS, atomp, rcstate, 3)) {
		return NULL;
	      }
	    }
	  } else {
	    if (!reginsert(isLazy?LREPEAT:REPEAT, atomp, rcstate, 6)) {
	      return NULL;
	    }
	    if (max == -1) {
	      max = 0;
	    }
	    OPERAND(atomp)[0] = min;
	    OPERAND(atomp)[1] = max;
	    OPERAND(atomp)[2] = (char)UNOPTIMIZED;
	  }
	} else {
	  if (min > 1) {
	    int i;
	    for (i = 1; i < min; i++) {
	      atomp = regcpnode(atomp, rcstate);
	      if (atomp == NULL) {
		return NULL;
	      }
	    }
	  }
	  if (min != max) {
	    if (max == -1) {
	      if (min == 0) {
		if (!reginsert(isLazy?LBRANCH:BRANCH, atomp, rcstate, 4)) {
		  return NULL;
		}
		OPERAND(atomp)[0] = (char)UNOPTIMIZED;
		regoptail(atomp, regnode(BACK,rcstate));
		regoptail(atomp, atomp);
	      } else {
		next = regnode(isLazy?LBRANCH:BRANCH,rcstate);
		regc(UNOPTIMIZED, rcstate);
		regtail(atomp, next);
		regtail(regnode(BACK,rcstate), atomp);
	      }
	    } else {
	      char* tailnode;
	      int optionalCount = max - min;
	      if (min > 0) {
		atomp = regcpnode(atomp, rcstate);
		if (atomp == NULL) {
		  return NULL;
		}
	      }
	      if (!reginsert(isLazy?LBRANCH:BRANCH, atomp, rcstate, 4)) {
		return NULL;
	      }
	      OPERAND(atomp)[0] = (char)UNOPTIMIZED;
	      while (optionalCount > 1) {
		char* previous = atomp;
		atomp = regcpnode(atomp, rcstate);
		if (atomp == NULL) {
		  return NULL;
		}
		regoptail(previous, atomp);
		optionalCount--;
	      }
	      tailnode = regnode(NOTHING, rcstate);
	      regoptail(atomp, tailnode);
	      regtail(atomp, tailnode);
	    }
	  }
	}

	if (rcstate->extflags & REGEXP_XMODE) {
	  eatWhiteSpace(rcstate);
	}
	if (rcstate->regparse != rcstate->regparseEnd &&
	    ISMULT(*rcstate->regparse)) {
	  if ((rcstate->extflags & REGEXP_PERLMODE) ||
	      *rcstate->regparse != '{')
	    FAIL("nested *?+{");
	    /* NOTREACHED */
	}

	return(ret);
}

static int getRangeChar(rcstate)
struct regcomp_state *rcstate;
{
  int res = UCHAR(*rcstate->regparse);
  rcstate->regparse++;
#ifdef META_ATSIGN
  if (rcstate->extflags & REGEXP_PERLMODE) {
    if (res == '@') {
      if (rcstate->regparse != rcstate->regparseEnd) {
	res = UCHAR(*rcstate->regparse);
	rcstate->regparse++;
	switch (res) {
	case 'a':
	  res = '\007';
	  break;
	case 'b':
	  res = '\b';
	  break;
	case 'f':
	  res = '\f';
	  break;
	case 'n':
	  res = '\n';
	  break;
	case 'r':
	  res = '\r';
	  break;
	case 't':
	  res = '\t';
	  break;
	case 'v':
	  res = '\v';
	  break;
	case 'x': {
	  unsigned long val = 0;
	  int found = 0;
	  while (rcstate->regparse != rcstate->regparseEnd &&
		 isxdigit(*rcstate->regparse)) {
	    found = 1;
	    val *= 16;
	    if (isdigit(*rcstate->regparse)) {
	      val += *rcstate->regparse - '0';
	    } else {
	      val += TOLC(*rcstate->regparse) - ('a' - 10);
	    }
	    rcstate->regparse++;
	  }
	  if (found) {
	    res = UCHAR(val);
	  }
	  break;
	}
	default: {
	  /* leave res as is */
	}
	}
      }
    }
  }
#endif
  return res;
}

/*
 - regatom - the lowest level
 *
 * Optimization:  gobbles an entire sequence of ordinary characters so that
 * it can turn them into a single node, which is smaller to store and
 * faster to run.  Backslashed characters are exceptions, each becoming a
 * separate node; the code is simpler that way and it's not worth fixing.
 */
static char *
regatom(flagp, rcstate, chain)
int *flagp;
struct regcomp_state *rcstate;
char *chain;
{
	register char *ret;
	int flags;
	int c;

	*flagp = WORST;		/* Tentatively. */

	if (rcstate->regparse == rcstate->regparseEnd) {
	  return regnode(NOTHING,rcstate);
	}

	c = *rcstate->regparse++;
	switch (c) {
	case '^':
	  if (rcstate->extflags & REGEXP_MLINEMODE) {
	    ret = regnode(BOL,rcstate);
	  } else {
	    ret = regnode(BOS,rcstate);
	  }
	  break;
	case '$':
	  if (rcstate->extflags & REGEXP_MLINEMODE) {
	    ret = regnode(EOL,rcstate);
	  } else {
	    ret = regnode(EOS,rcstate);
	  }
	  break;
	case '.':
	  if (rcstate->extflags & REGEXP_MLINEMODE) {
	    ret = regnode(ISNOTCHAR,rcstate);
	    regc('\n', rcstate);
	  } else {
	    ret = regnode(ANY,rcstate);
	  }
	  *flagp |= HASWIDTH|SIMPLE;
	  break;
	case '[':
	  ret = parseCharRange(rcstate, ']');
	  if (ret == NULL) {
	    return(NULL);
	  }
	  *flagp |= HASWIDTH|SIMPLE;
	  break;
	case '(': {
	        int parenFlags = GROUP_BIT|CAPTURE_BIT;
	        if (rcstate->regparse != rcstate->regparseEnd &&
		    rcstate->regparse[0] == '?') {
		  rcstate->regparse++; /* eat the '?' */
		  ret = parseQuestionMod(flagp, rcstate, &parenFlags, chain);
		  if (ret != NULL || parenFlags == 0) {
		    if (ret) {
		      if (rcstate->extflags & REGEXP_XMODE) {
			eatWhiteSpace(rcstate);
		      }
		    }
		    return (ret);
		  }
		  /* fall through to process regexp in the parens */
		}
		ret = reg(parenFlags, &flags, rcstate);
		if (ret == NULL) {
		  return(NULL);
		}
		*flagp |= flags;
		break;
	}
	case '|':
	case ')':
		FAIL("internal urp");	/* Supposed to be caught earlier. */
		/* NOTREACHED */
	case '?':
		FAIL("? follows nothing");
		/* NOTREACHED */
	case '+':
		FAIL("+ follows nothing");
		/* NOTREACHED */
	case '*':
		FAIL("* follows nothing");
		/* NOTREACHED */
	case '\\':
	        if (rcstate->regparse == rcstate->regparseEnd) {
		  FAIL("nothing follows \"\\\"");
		  /* NOTREACHED */
		}
#ifndef META_ATSIGN
		if (rcstate->extflags & REGEXP_PERLMODE) {
		  ret = parseMacroMod(flagp, rcstate, '\\');
		} else
#endif
		{
		  if (rcstate->extflags & REGEXP_NOCASE) {
		    ret = regnode(NCEXACTLY,rcstate);
		  } else {
		    ret = regnode(EXACTLY,rcstate);
		  }
		  regc(0, rcstate);
		  regc(1, rcstate);
		  if (rcstate->extflags & REGEXP_NOCASE) {
		    regc(TOLC(*rcstate->regparse++),rcstate);
		  } else {
		    regc(*rcstate->regparse++,rcstate);
		  }
		  *flagp |= HASWIDTH|SIMPLE;
		}
		break;
	case '{':
#ifdef META_ATSIGN
	case '@':
	  if (rcstate->extflags & REGEXP_PERLMODE) {
	    if (c == '@') {
	      if (rcstate->regparse == rcstate->regparseEnd) {
		FAIL("nothing follows @");
		/* NOTREACHED */
	      }
	      ret = parseMacroMod(flagp, rcstate, '@');
	      break;
	    } else {
	      FAIL("{ follows nothing");
	      /* NOTREACHED */
	    }
	  }
#else
	  if (rcstate->extflags & REGEXP_PERLMODE) {
	      FAIL("{ follows nothing");
	      /* NOTREACHED */
	  }
#endif
	  /* FALL THROUGH to default */
	default: {
		  int len;
		  char ender;
		  CONST char *endp;

		  rcstate->regparse--;
		  endp = rcstate->regparse;
		  if (rcstate->extflags & REGEXP_XMODE) {
		    while (endp != rcstate->regparseEnd && !ISMETAWHITE(*endp)) {
		      if ((rcstate->extflags & REGEXP_PERLMODE) &&
			  (*endp == '{'
#ifdef META_ATSIGN
			   || *endp == '@'
#endif
			  )) {
			break;
		      }
		      endp++;
		    }
		  } else {
		    while (endp != rcstate->regparseEnd && !ISMETA(*endp)) {
		      if ((rcstate->extflags & REGEXP_PERLMODE) &&
			  (*endp == '{'
#ifdef META_ATSIGN
			   || *endp == '@'
#endif
			  )) {
			break;
		      }
		      endp++;
		    }
		  }
		  len = endp - rcstate->regparse;
		  if (len <= 0) {
		    FAIL("internal disaster");
		    /* NOTREACHED */
		  }
		  ender = rcstate->regparse[len];
		  if (len > 1 && ISMULT(ender)) {
		    if (ender != '{' || (rcstate->extflags & REGEXP_PERLMODE))
		      len--;  /* Back off clear of ?+*{ operand. */
		  }
		  *flagp |= HASWIDTH;
		  if (len == 1)
		    *flagp |= SIMPLE;
		  if (rcstate->extflags & REGEXP_NOCASE) {
		    ret = regnode(NCEXACTLY,rcstate);
		  } else {
		    ret = regnode(EXACTLY,rcstate);
		  }
		  regc(len >> 8, rcstate);
		  regc(len & 0x00ff, rcstate);
		  if (rcstate->extflags & REGEXP_NOCASE) {
		    while (len > 0) {
		      regc(TOLC(*rcstate->regparse++),rcstate);
		      len--;
		    }
		  } else {
		    while (len > 0) {
		      regc(*rcstate->regparse++,rcstate);
		      len--;
		    }
		  }
		}
		break;
	}

	if (rcstate->extflags & REGEXP_XMODE) {
	  eatWhiteSpace(rcstate);
	}
	return(ret);
}


static void
addcs(csp, c)
struct re_classset *csp; 
int c;
{
  if (csp->csset[UCHAR(c)] == csp->negated) {
    if (csp->negated) {
      csp->csset[UCHAR(c)] = 0;
      csp->cstotal -= UCHAR(c);
      csp->cscount--;
    } else {
      csp->csset[UCHAR(c)] = 1;
      csp->cstotal += UCHAR(c);
      csp->cscount++;
    }
  }
}

static void
addclasschar(c, rcstate, csp)
int c;
struct regcomp_state *rcstate;
struct re_classset *csp;
{
  addcs(csp, c);
  if (rcstate->extflags & REGEXP_NOCASE) {
    if (islower(c)) {
      addcs(csp, toupper(c));
    } else if (isupper(c)) {
      addcs(csp, tolower(c));
    }
  }
}

/* Returns 1 if c is not already in csp.
 * Return 0 if c is in csp.
 */
static int
addclassnotchar(c, csp)
int c;
struct re_classset *csp;
{
  if (csp->csset[UCHAR(c)]) {
    return 0;
  } else {
    int i;
    for (i = 0; i < UCHAR(c); i++) {
      addcs(csp, i);
    }
    for (i = UCHAR(c)+1; i < CHARMAP_SIZE; i++) {
      addcs(csp, i);
    }
    return 1;
  }
}

static void
initcs(csp, negated)
struct re_classset *csp;
int negated;
{
  csp->cscount = 0;
  csp->cstotal = 0;
  csp->negated = negated;
  if (negated) {
    int i;
    csp->cscount = CHARMAP_SIZE;
    if (CHARMAP_SIZE == 256) {
      csp->cstotal = MAX_CSTOTAL;
    } else {
      for (i = 1; i < CHARMAP_SIZE; i++) {
	csp->cstotal += i;
      }
    }
  }
  memset(csp->csset, negated, sizeof(csp->csset));
}

static void
initBuiltinClass(csp, fptr, negated)
struct re_classset *csp; 
re_classtestProc *fptr;
int negated;
{
  int i;
  initcs(csp, negated);
  for (i=0; i < sizeof(csp->csset); i++) {
    if (fptr(i)) {
      addcs(csp, i);
    }
  }
}

static char biIntersectTable[CS_LAST][CS_LAST];

static int doSetsIntersect(csp1, csp2)
char *csp1;
char *csp2;
{
  int i;
  /* Note we treat csset as an array of longs for speed */
  long *p1 = (long*)csp1;
  long *p2 = (long*)csp2;
  for (i = 0; i < (CHARMAP_SIZE / sizeof(long)); i++) {
    if (*p1++ & *p2++) {
      /* Found something they both have in common. */
      return 1;
    }
  }
  return 0;
}

/* Following wrapper functions exist because the is* ctype guys
 * are not functions on all platforms.
 */
static int re_isspace(c)
int c;
{
  return isspace(c);
}

static int re_isdigit(c)
int c;
{
  return isdigit(c);
}

static int re_isalnum(c)
int c;
{
  return isalnum(c);
}

static int re_isalpha(c)
int c;
{
  return isalpha(c);
}

static int re_isupper(c)
int c;
{
  return isupper(c);
}

static int re_islower(c)
int c;
{
  return islower(c);
}

static int re_isxdigit(c)
int c;
{
  return isxdigit(c);
}

static int re_ispunct(c)
int c;
{
  return ispunct(c);
}

static int re_isgraph(c)
int c;
{
  return isgraph(c);
}

static int re_isprint(c)
int c;
{
  return isprint(c);
}

static int re_iscntrl(c)
int c;
{
  return iscntrl(c);
}

static void
initBuiltinClasses()
{
  CONST char* p;
  int i, j;
  if (builtin_initialized) {
    return;
  }
  builtin_initialized = 1;

  for (i=0; i < sizeof(lcmap); i++) {
    if (isupper(i)) {
      lcmap[i] = (char)tolower(i);
    } else {
      lcmap[i] = i;
    }
  }

  initBuiltinClass(&bitab[CS_SPACE], re_isspace, 0);
  initBuiltinClass(&bitab[CS_NOTSPACE], re_isspace, 1);
  initBuiltinClass(&bitab[CS_DIGIT], re_isdigit, 0);
  initBuiltinClass(&bitab[CS_NOTDIGIT], re_isdigit, 1);
  initBuiltinClass(&bitab[CS_WORD], re_isalnum, 0);
  addcs(&bitab[CS_WORD], '_');
  initBuiltinClass(&bitab[CS_NOTWORD], re_isalnum, 1);
  addcs(&bitab[CS_NOTWORD], '_');
  initBuiltinClass(&bitab[CS_UPPER], re_isupper, 0);
  initBuiltinClass(&bitab[CS_NOTUPPER], re_isupper, 1);
  initBuiltinClass(&bitab[CS_LOWER], re_islower, 0);
  initBuiltinClass(&bitab[CS_NOTLOWER], re_islower, 1);
  initBuiltinClass(&bitab[CS_ALPHA], re_isalpha, 0);
  initBuiltinClass(&bitab[CS_NOTALPHA], re_isalpha, 1);
  initBuiltinClass(&bitab[CS_ALPHANUM], re_isalnum, 0);
  initBuiltinClass(&bitab[CS_NOTALPHANUM], re_isalnum, 1);
  initBuiltinClass(&bitab[CS_XDIGIT], re_isxdigit, 0);
  initBuiltinClass(&bitab[CS_NOTXDIGIT], re_isxdigit, 1);
  initBuiltinClass(&bitab[CS_PUNCT], re_ispunct, 0);
  initBuiltinClass(&bitab[CS_NOTPUNCT], re_ispunct, 1);
  initBuiltinClass(&bitab[CS_GRAPH], re_isgraph, 0);
  initBuiltinClass(&bitab[CS_NOTGRAPH], re_isgraph, 1);
  initBuiltinClass(&bitab[CS_PRINT], re_isprint, 0);
  initBuiltinClass(&bitab[CS_NOTPRINT], re_isprint, 1);
  initBuiltinClass(&bitab[CS_CNTRL], re_iscntrl, 0);
  initBuiltinClass(&bitab[CS_NOTCNTRL], re_iscntrl, 1);

  for (i = 0; i < CS_LAST; i++) {
    for (j = 0; j < CS_LAST; j++) {
      biIntersectTable[i][j] = doSetsIntersect(bitab[i].csset,
					       bitab[j].csset);
    }
  }

  initcs(&csMult, 0);
  p = MULT;
  while (*p != 0) {
    addcs(&csMult, *p++);
  }

  initcs(&csMeta, 0);
  p = META;
  while (*p != 0) {
    addcs(&csMeta, *p++);
  }

  initcs(&csMetawhite, 0);
  p = METAWHITE;
  while (*p != 0) {
    addcs(&csMetawhite, *p++);
  }
}

static char*
addclassnode(rcstate, negated, csp)
struct regcomp_state *rcstate;
int negated;
struct re_classset *csp;
{
  int i;
  char* ret;

  if (csp->cscount == 1) {
    /* Just one character in the set so no need for class */
    ret = regnode(EXACTLY,rcstate);
    regc(0, rcstate);
    regc(1, rcstate);
    regc(csp->cstotal, rcstate);
    return ret;
  }
  if (csp->cscount == CHARMAP_SIZE-1) {
    /* Just one character not in the set so no need for class */
    ret = regnode(ISNOTCHAR,rcstate);
    regc(MAX_CSTOTAL - csp->cstotal, rcstate);
    return ret;
  }
  if (csp->cscount == CHARMAP_SIZE) {
    /* All characters */
    return regnode(ANY,rcstate);
  }

  /* See if it matches one that is already created */
  for (i=0; i < rcstate->regnClassCount; i++) {
    if (csp->cscount == rcstate->classtab[i].cscount &&
	csp->cstotal == rcstate->classtab[i].cstotal) {
      if (memcmp(csp->csset, rcstate->classtab[i].csset,
		 sizeof(csp->csset)) == 0) {
	/* found one that matches */
	ret = regnode(ANYOF, rcstate);
	regc(i, rcstate);
	return ret;
      }
    }
  }

  /* look for a match with one of the builtin classes */
  for (i=0; i < CS_LAST; i++) {
    if (csp->cscount == bitab[i].cscount &&
	csp->cstotal == bitab[i].cstotal) {
      if (memcmp(csp->csset, bitab[i].csset, sizeof(csp->csset)) == 0) {
	/* found a builtin that matches */
	ret = regnode(BIANYOF,rcstate);
	regc(i, rcstate);
	return ret;
      }
    }
  }

  /* Couldn't find one so add a new one */
  ret = regnode(ANYOF, rcstate);
  regc(rcstate->regnClassCount, rcstate);
  rcstate->regnClassCount++;
  return ret;
}

static void
eatWhiteSpace(rcstate)
struct regcomp_state *rcstate;
{
  if (rcstate->regparse == rcstate->regparseEnd ||
      !BITEST(CS_SPACE, rcstate->regparse[0])) {
    return;
  }
  for (;;) {
    do {
      rcstate->regparse++;
      if (rcstate->regparse == rcstate->regparseEnd) {
	return;
      }
    } while (BITEST(CS_SPACE, rcstate->regparse[0]));
    if (*rcstate->regparse != '#') {
      return;
    }
    /* parse the comment */
    do {
      rcstate->regparse++;
      if (rcstate->regparse == rcstate->regparseEnd) {
	return;
      }
    } while (*rcstate->regparse != '\n');
  }
}

static char*
parseQuestionMod(flagp, rcstate, parenFlagsPtr, chain)
int *flagp;
struct regcomp_state *rcstate;
int *parenFlagsPtr;
char *chain;
{
  char* ret;
  if (rcstate->regparse == rcstate->regparseEnd) {
    *parenFlagsPtr = 0; /* lets the caller know we have an error */
    FAIL("invalid character after (?");
    /* NOTREACHED */
  }
    
  switch (*rcstate->regparse++) {
  case ':':
    *parenFlagsPtr = GROUP_BIT;
    return NULL;
  case '=':
    *parenFlagsPtr = GROUP_BIT|LOOK_BIT;
    return NULL;
  case '!':
    *parenFlagsPtr = GROUP_BIT|NLOOK_BIT;
    return NULL;
  case '|': {
    /* What follows is a character range. (?|...) is an
     * alternate syntax for [...] which has the benefit
     * of not using any special Tcl characters that need
     * to be escaped.
     */
    ret = parseCharRange(rcstate, ')');
    if (ret != NULL) {
      *flagp |= HASWIDTH|SIMPLE;
    } else {
      *parenFlagsPtr = 0; /* we have an error */
    }
    return ret;
  }
  case '#':
    while (rcstate->regparse != rcstate->regparseEnd &&
	   *rcstate->regparse != ')') {
      rcstate->regparse++;
    }
    if (rcstate->regparse == rcstate->regparseEnd) {
      *parenFlagsPtr = 0; /* lets the caller know we have an error */
      FAIL("missing ')' for (?# comment");
      /* NOTREACHED */
    }
    rcstate->regparse++; /* eat the ')' */
    /* All done parsing (?#...)
     * A comment is not a token so return chain if it exists or get an atom.
     */
    if (chain) {
      return chain;
    } else {
      if (rcstate->extflags & REGEXP_XMODE) {
	eatWhiteSpace(rcstate);
      }
      ret = regatom(flagp, rcstate, chain);
      if (ret == NULL) {
	*parenFlagsPtr = 0; /* we have an error */
      }
      return ret;
    }
  case 'i':
    if (!(rcstate->extflags & REGEXP_NOCASE)) {
      *parenFlagsPtr = RECOMPILE_BIT;
    }
    rcstate->extflags |= REGEXP_NOCASE;
    return parseQuestionMod(flagp, rcstate, parenFlagsPtr, chain);
  case 'm':
    if (!(rcstate->extflags & REGEXP_MLINEMODE)) {
      *parenFlagsPtr = RECOMPILE_BIT;
    }
    rcstate->extflags |= REGEXP_MLINEMODE;
    return parseQuestionMod(flagp, rcstate, parenFlagsPtr, chain);
  case 'p':
    if (!(rcstate->extflags & REGEXP_PERLMODE)) {
      *parenFlagsPtr = RECOMPILE_BIT;
    }
    rcstate->extflags |= REGEXP_PERLMODE;
    return parseQuestionMod(flagp, rcstate, parenFlagsPtr, chain);
  case 'q':
    if (!(rcstate->extflags & REGEXP_QUICKCOMPILE)) {
      *parenFlagsPtr = RECOMPILE_BIT;
    }
    rcstate->extflags |= REGEXP_QUICKCOMPILE;
    return parseQuestionMod(flagp, rcstate, parenFlagsPtr, chain);
  case 'x':
    if (!(rcstate->extflags & REGEXP_XMODE)) {
      *parenFlagsPtr = RECOMPILE_BIT;
    }
    rcstate->extflags |= REGEXP_XMODE;
    return parseQuestionMod(flagp, rcstate, parenFlagsPtr, chain);
  case ')':
    /* All done parsing options (?...)
     * Options are not a token so return chain if it exists or get an atom.
     */
    if (*parenFlagsPtr == RECOMPILE_BIT) {
      *parenFlagsPtr = 0;
      return NULL;
    }
    if (chain) {
      return chain;
    } else {
      if (rcstate->extflags & REGEXP_XMODE) {
	eatWhiteSpace(rcstate);
      }
      ret = regatom(flagp, rcstate, chain);
      if (ret == NULL) {
	*parenFlagsPtr = 0; /* we have an error */
      }
      return ret;
    }
  default:
    *parenFlagsPtr = 0; /* lets the caller know we have an error */
    FAIL("invalid character after (?");
    /* NOTREACHED */
  }
}

static char* parseCharRange(rcstate, termChar)
struct regcomp_state *rcstate;
int termChar;
{
#define INVALID_RANGE_CHAR 256
  int clss = INVALID_RANGE_CHAR;
  int classend;
  int negated = 0;
  struct re_classset* cs;
  int ok;

  if (rcstate->regnClassCount == MAX_CLASS_COUNT) {
    FAIL("too many [...] ranges");
    /* NOTREACHED */
  }
  cs = &rcstate->classtab[rcstate->regnClassCount];
  if (rcstate->regparse != rcstate->regparseEnd &&
      *rcstate->regparse == '^') { /* Complement of range. */
    negated = 1;
    rcstate->regparse++;
  }
  initcs(cs, negated);
  if (rcstate->regparse != rcstate->regparseEnd &&
      (*rcstate->regparse == termChar || *rcstate->regparse == '-')) {
    clss = *rcstate->regparse++;
    addclasschar(clss, rcstate, cs);
  }
  while (rcstate->regparse != rcstate->regparseEnd &&
	 *rcstate->regparse != termChar) {
    int nc = getRangeChar(rcstate);
    if (nc == '-') {
      if (rcstate->regparse == rcstate->regparseEnd ||
	  *rcstate->regparse == termChar) {
	addclasschar('-', rcstate, cs);
      } else {
	classend = getRangeChar(rcstate);
	if (clss > classend) {
	  if (termChar == ']') {
	    FAIL("invalid [] range");
	  } else {
	    FAIL("invalid (?|) range");
	  }
	  /* NOTREACHED */
	}
	for (; clss <= classend; clss++) {
	  addclasschar(clss, rcstate, cs);
	}
	clss = classend;
      }
    } else if (nc == '[' && getPosixRangeClass(rcstate, cs, &ok)) {
      clss = INVALID_RANGE_CHAR;
      if (!ok) {
	FAIL("unknown name in class [:name:]");
      }
    } else {
      clss = nc;
      addclasschar(clss, rcstate, cs);
    }
  }
  if (rcstate->regparse == rcstate->regparseEnd ||
      *rcstate->regparse != termChar) {
    if (termChar == ']') {
      FAIL("unmatched []");
    } else {
      FAIL("unmatched (?|)");
    }
    /* NOTREACHED */
  }
  rcstate->regparse++;
  return addclassnode(rcstate, negated, cs);
}

typedef struct {
  CONST char* name;
  int csidx;
} posixClassStruct;

static posixClassStruct posixTable[] = {
  {"space", CS_SPACE},
  {"alpha", CS_ALPHA},
  {"alnum", CS_ALPHANUM},
  {"upper", CS_UPPER},
  {"lower", CS_LOWER},
  {"digit", CS_DIGIT},
  {"xdigit", CS_XDIGIT},
  {"punct", CS_PUNCT},
  {"graph", CS_GRAPH},
  {"print", CS_PRINT},
  {"cntrl", CS_CNTRL},
  {NULL, -1}
};

static int 
getPosixRangeClass(rcstate, csp, ok)
struct regcomp_state *rcstate;
struct re_classset *csp;
int *ok;
{
  int i = 0;
  int namelen = 0;
  char *nameptr;
  char *p = rcstate->regparse;
  *ok = 1;

  if (p == rcstate->regparseEnd || p[0] != ':') {
    return 0;
  }
  p++;
  nameptr = p;
  while (p != rcstate->regparseEnd && isalpha(*p)) {
    namelen++;
    p++;
  }
  if (p == rcstate->regparseEnd || p[0] != ':') {
    return 0;
  }
  p++;
  if (p == rcstate->regparseEnd || p[0] != ']') {
    return 0;
  }
  p++;
  rcstate->regparse = p;
  while (posixTable[i].name != NULL) {
    if ((int)strlen(posixTable[i].name) == namelen) {
      if (memcmp(posixTable[i].name, nameptr, namelen) == 0) {
	charMapAdd(csp, &bitab[posixTable[i].csidx]);
	return 1;
      }
    }
    i++;
  }
  *ok = 0;
  return 1;
}

static char*
parseMacroMod(flagp, rcstate, macroOp)
int *flagp;
struct regcomp_state *rcstate;
int macroOp;
{
  char* ret;
  char* lenPtr;
  int len = 0;
  /* Note: caller already checked for EOF */
  char macroArg = *rcstate->regparse++;

  switch (macroArg) {
  case 'A':
    ret = regnode(BOS,rcstate);
    break;
  case 'Z':
    ret = regnode(EOS,rcstate);
    break;
  case 'G':
    ret = regnode(BOINPUT,rcstate);
    break;
  case 'b':
    ret = regnode(WBOUND,rcstate);
    break;
  case 'B':
    ret = regnode(NWBOUND,rcstate);
    break;
  case '<':
    ret = regnode(SOW,rcstate);
    break;
  case '>':
    ret = regnode(EOW,rcstate);
    break;
  case 'd':
    ret = regnode(BIANYOF,rcstate);
    regc(CS_DIGIT, rcstate);
    *flagp |= HASWIDTH|SIMPLE;
    break;
  case 'D':
    ret = regnode(BIANYOF,rcstate);
    regc(CS_NOTDIGIT, rcstate);
    *flagp |= HASWIDTH|SIMPLE;
    break;
  case 's':
    ret = regnode(BIANYOF,rcstate);
    regc(CS_SPACE, rcstate);
    *flagp |= HASWIDTH|SIMPLE;
    break;
  case 'S':
    ret = regnode(BIANYOF,rcstate);
    regc(CS_NOTSPACE, rcstate);
    *flagp |= HASWIDTH|SIMPLE;
    break;
  case 'w':
    ret = regnode(BIANYOF,rcstate);
    regc(CS_WORD, rcstate);
    *flagp |= HASWIDTH|SIMPLE;
    break;
  case 'W':
    ret = regnode(BIANYOF,rcstate);
    regc(CS_NOTWORD, rcstate);
    *flagp |= HASWIDTH|SIMPLE;
    break;
  case '1':
  case '2':
  case '3':
  case '4':
  case '5':
  case '6':
  case '7':
  case '8':
  case '9': {
    int brNum = macroArg - '0';
    if (rcstate->regnclosepar > 9) {
      if (rcstate->regparse != rcstate->regparseEnd &&
	  *rcstate->regparse >= '0' && *rcstate->regparse <= '9') {
	brNum *= 10;
	brNum += *rcstate->regparse - '0';
	rcstate->regparse++;
	if (rcstate->regnclosepar > 99) {
	  if (rcstate->regparse != rcstate->regparseEnd &&
	      *rcstate->regparse >= '0' && *rcstate->regparse <= '9') {
	    brNum *= 10;
	    brNum += *rcstate->regparse - '0';
	    rcstate->regparse++;
	  }
	}
      }
    }
    if (brNum > rcstate->regnclosepar) {
      FAIL("no capture parentheses for backreference");
      /* NOTREACHED */
    }
    ret = regnode(BACKREF,rcstate);
    regc(brNum, rcstate);
    if (rcstate->regparhaswidth[brNum]) {
      *flagp |= HASWIDTH;
    }
    break;
  }
  case 'Q':
    len = 0;
    /* scan for "\E" or end of input */
    if (rcstate->regparse == rcstate->regparseEnd) {
      ret = regnode(NOTHING,rcstate);
      break;
    }
    if (rcstate->regparse[0] == macroOp &&
	((rcstate->regparse+1) != rcstate->regparseEnd &&
	 rcstate->regparse[1] == 'E')) {
      rcstate->regparse += 2;
      ret = regnode(NOTHING,rcstate);
      break;
    }
    if (rcstate->extflags & REGEXP_NOCASE) {
      ret = regnode(NCEXACTLY,rcstate);
    } else {
      ret = regnode(EXACTLY,rcstate);
    }
    /* save length fields */
    lenPtr = rcstate->regcode;
    regc(0, rcstate);
    regc(0, rcstate);
    do {
      len++;
      if (rcstate->extflags & REGEXP_NOCASE) {
	regc(TOLC(*rcstate->regparse++),rcstate);
      } else {
	regc(*rcstate->regparse++,rcstate);
      }
    } while (rcstate->regparse != rcstate->regparseEnd &&
	     (rcstate->regparse[0] != macroOp ||
	      ((rcstate->regparse+1) == rcstate->regparseEnd ||
	       rcstate->regparse[1] != 'E')));
    /* Adjust regparse ptr to eat the terminator. */
    if (rcstate->regparse != rcstate->regparseEnd) {
      rcstate->regparse += 2;
    }
    lenPtr[0] = len >> 8;
    lenPtr[1] = len & 0x00ff;
    *flagp |= HASWIDTH;
    if (len == 1) {
      *flagp |= SIMPLE;
    }
    break;
#ifdef META_ATSIGN
  case 'a':
    ret = regnode(EXACTLY,rcstate);
    regc(0, rcstate);
    regc(1, rcstate);
    regc('\007', rcstate);
    *flagp |= HASWIDTH|SIMPLE;
    break;
  case 'f':
    ret = regnode(EXACTLY,rcstate);
    regc(0, rcstate);
    regc(1, rcstate);
    regc('\f', rcstate);
    *flagp |= HASWIDTH|SIMPLE;
    break;
  case 'n':
    ret = regnode(EXACTLY,rcstate);
    regc(0, rcstate);
    regc(1, rcstate);
    regc('\n', rcstate);
    *flagp |= HASWIDTH|SIMPLE;
    break;
  case 'r':
    ret = regnode(EXACTLY,rcstate);
    regc(0, rcstate);
    regc(1, rcstate);
    regc('\r', rcstate);
    *flagp |= HASWIDTH|SIMPLE;
    break;
  case 't':
    ret = regnode(EXACTLY,rcstate);
    regc(0, rcstate);
    regc(1, rcstate);
    regc('\t', rcstate);
    *flagp |= HASWIDTH|SIMPLE;
    break;
  case 'v':
    ret = regnode(EXACTLY,rcstate);
    regc(0, rcstate);
    regc(1, rcstate);
    regc('\v', rcstate);
    *flagp |= HASWIDTH|SIMPLE;
    break;
#endif
#ifdef META_ATSIGN
  case '@':
#endif
  case '\\':
  case '*':
  case '+':
  case '?':
  case '(':
  case ')':
  case '|':
  case '[':
  case ']':
  case '{':
  case '}':
  case '^':
  case '$':
  case '.':
    ret = regnode(EXACTLY,rcstate);
    regc(0, rcstate);
    regc(1, rcstate);
    regc(macroArg, rcstate);
    *flagp |= HASWIDTH|SIMPLE;
    break;
  default: {
#if 0
      char msg[80];
      sprintf(msg, "unexpected '%c' found after %c", macroArg, macroOp);
      FAIL(msg);
#else
#ifdef META_ATSIGN
      FAIL("unexpected character after '@'");
#else
      FAIL("unexpected character after \"\\\"");
#endif
#endif
      /* NOTREACHED */
    }
    break;
  }
  return (ret);
}

/* Delete 'node' by moving everything after it down.
 * Need to be careful to only do this if a next does
 * not 'jump' over the node being deleted.
 */
static void
regrmnode(node, rcstate, size)
char *node;
struct regcomp_state *rcstate;
int size;
{
  size_t len = rcstate->regcode - (node+size);
  rcstate->regcode -= size;
  if (len > 0) {
#ifdef HAS_MEMMOVE
    memmove(node, node+size, len);
#else
    char* src = node+size;
    char* dst = node;
    char* end = src+len;
    while (src < end) {
      *dst++ = *src++;
    }
#endif
  }
}

/* Makes a copy from node to the end of whats been compiled so far.
 * It then appends the copy onto the end of the compiled code and
 * links 'nodes' tail to the copy.
 * Returns a pointer to the new copy.
 */
static char*
regcpnode(node, rcstate)
char *node;
struct regcomp_state *rcstate;
{
  char* res = rcstate->regcode;
  size_t len = res - node;
  if (rcstate->regcode+len >= rcstate->regcodeEnd) {
    FAIL("regexp too big");
    /* NOTREACHED */
  }
  rcstate->regcode += len;
  memcpy(res, node, len);
  regtail(node, res);
  return res;
}

/*
 - regnode - emit a node
 */
static char *			/* Location. */
regnode(op, rcstate)
int op;
struct regcomp_state *rcstate;
{
	register char *ret;
	register char *ptr;

	ret = rcstate->regcode;
	
	if (rcstate->regcode+3 >= rcstate->regcodeEnd) {
	  FAIL("regexp too big");
	  /* NOTREACHED */
	}

	ptr = ret;
	*ptr++ = (char)op;
	*ptr++ = '\0';		/* Null "next" pointer. */
	*ptr++ = '\0';
	rcstate->regcode = ptr;

	return(ret);
}

/*
 - regc - emit (if appropriate) a byte of code
 */
static void
regc(b, rcstate)
int b;
struct regcomp_state *rcstate;
{
  if (rcstate->regcode < rcstate->regcodeEnd) {
    *rcstate->regcode++ = (char)b;
  }
#if 0
  else {
    /* Let someone else catch this error.
     * The regnode that adds the END will catch
     * that we ran out of space.
     */
  }
#endif
}

/*
 - reginsert - insert an operator in front of already-emitted operand
 *
 * Means relocating the operand.
 */
static char*
reginsert(op, opnd, rcstate, size)
int op;
char *opnd;
struct regcomp_state *rcstate;
int size;
{
  if (rcstate->regcode+size >= rcstate->regcodeEnd) {
    FAIL("regexp too big");
    /* NOTREACHED */
  }
#ifdef HAS_MEMMOVE
  memmove(opnd+size, opnd, rcstate->regcode - opnd);
  rcstate->regcode += size;
#else
  {
    char *dst;
    char *src = rcstate->regcode;
    rcstate->regcode += size;
    dst = rcstate->regcode;
    while (src > opnd) {
      *--dst = *--src;
    }
  }
#endif

  *opnd++ = (char)op;
  while (size > 1) {
    *opnd++ = '\0';
    size--;
  }
  return opnd;
}

/*
 - regtail - set the next-pointer at the end of a node chain
 */
static void
regtail(p, val)
char *p;
char *val;
{
	register char *scan;
	register char *temp;
	register int offset;

	/* Find last node. */
	scan = p;
	for (;;) {
		temp = regnext(scan);
		if (temp == NULL)
			break;
		scan = temp;
	}

	if (OP(scan) == BACK)
		offset = scan - val;
	else
		offset = val - scan;
	*(scan+1) = (char)((offset>>8)&0377);
	*(scan+2) = (char)(offset&0377);
}

static void
regoptail(p, val)
char *p;
char *val;
{
  /* "Operandless" and "op != BRANCH" are synonymous in practice.
   */
  if (p != NULL) {
    switch (OP(p)) {
    case ORSTART:
    case BRANCH:
    case LBRANCH:
      regtail(OPERAND(p+1), val);
      break;
    case ORBRANCH:
    case OREND:
      regtail(OPERAND(p), val);
      break;
    default:
      break;
    }
  }
}

static char *findStartNode(next)
char *next;
{
  char op;
  for (;;) {
    if (next == NULL) {
      return NULL;
    }

    op = OP(next);
    switch (op) {
    case BOS:
    case EOS:
    case ANYOF:
    case NCEXACTLY:
    case EXACTLY:
    case BIANYOF:
    case BOL:
    case EOL:
    case BOINPUT:
    case ISNOTCHAR:
      /* We found a definite node so return it. */
      return next;

    case WBOUND:
    case NWBOUND:
    case SOW:
    case EOW: {
      /* We found a definite node but look ahead some more
       * to see if a better one can be found.
       * We can do this since these are just assertions.
       */
      char *nextsn = findStartNode(regoptnext(next));
      if (nextsn) {
	next = nextsn;
      }
      return next;
    }

    case ORSTART:
    case STAR:
    case LSTAR:
    case BRANCH:
    case LBRANCH:
      if (UCHAR(*OPERAND(next)) != UNOPTIMIZED) {
	return next;
      } else {
	return NULL;
      }

    case LOOK:
    case PLUS:
    case LPLUS:
      next = OPERAND(next);
      break;

    case REPEAT:
    case LREPEAT:
      if (OPERAND(next)[0] != 0) {
	/* repeats at least once so return the repeated node */
	next = OPERAND(next)+3;
	break;
      } else if (UCHAR(OPERAND(next)[2]) != UNOPTIMIZED) {
	return next;
      } else {
	/* next could be more than one thing so return NULL */
	return NULL;
      }

    case FASTLOOK:
    case NOTHING:
    case ENDLOOK:
    case OPEN:
    case CLOSE:
      /* try the next one */
      next = regoptnext(next);
      break;

    case ANY:
      /* ANY is not a good start node */
    case END:
    case NLOOK:
    case FASTNLOOK:
    case ORBRANCH:
    case OREND:
    case BACK:
    default:
      /* Could not find a good start node */
      return NULL;
    }
  }
}

/* Add the chars in 'src' to 'dst'.
 */
static void charMapAdd(dst, src)
struct re_classset* dst;
CONST struct re_classset* src;
{
  int i;
  int count = src->cscount;
  CONST char* srcbytes = src->csset;
  if (count == 0) {
    return;
  }
  for (i = 0; i < CHARMAP_SIZE; i++) {
    if (srcbytes[i]) {
      addcs(dst, i);
      count--;
      if (count == 0) {
	break;
      }
    }
  }
}

/* calcNextEatenByte will figure out if the branch pointed to be
 * 'next' will definitely eat a byte when it matches.
 * If it will eat a byte then the possible values of that byte will
 * be set in 'cmap'.
 * Otherwise returns 0 and leaves cmap alone.
 * NOTE: 'cmap' must point to an array of ChARMAP_SIZE chars
 * and must be a legal 'long*'.
 * Returns 1 if it could figure out what would be eaten next.
 * Returns 0 if it could not.
 */
static int calcNextEatenByte(rcstate, next, cmap)
struct regcomp_state *rcstate;
char *next;
struct re_classset *cmap;
{
  char op;
  for (;;) {
    if (next == NULL) {
      return 0;
    }

    op = OP(next);
    switch (op) {
    case EXACTLY:
    case NCEXACTLY:
      addclasschar(OPERAND(next)[2], rcstate, cmap);
      return 1;

    case ISNOTCHAR:
      return addclassnotchar(OPERAND(next)[0], cmap);

    case SOW:
      /* SOW is somewhat general so lookahead a bit more */
      if (!calcNextEatenByte(rcstate, regoptnext(next), cmap)) {
	/* Could not find anything more definite */
	charMapAdd(cmap, &bitab[CS_WORD]);
      }
      return 1;

    case EOW:
      /* EOW is somewhat general so lookahead a bit more */
      if (!calcNextEatenByte(rcstate, regoptnext(next), cmap)) {
	/* EOW will match end of input which does not eat a byte so return 0 */
	return 0;
      }
      return 1;

    case BRANCH:
    case LBRANCH: {
      int idx = UCHAR(*OPERAND(next));
      if (idx == UNOPTIMIZED) {
	if (!calcNextEatenByte(rcstate, regoptnext(next), cmap)) {
	  return 0;
	}
	next = OPERAND(next+1);
	break;
      } else if (idx < CS_LAST) {
	charMapAdd(cmap, &bitab[idx]);
	return 1;
      } else {
	charMapAdd(cmap, &rcstate->classtab[idx-CS_LAST]);
	return 1;
      }
    }

    case ORSTART: {
      int idx = UCHAR(*OPERAND(next));
      if (idx == UNOPTIMIZED) {
	return 0;
      } else if (idx < CS_LAST) {
	charMapAdd(cmap, &bitab[idx]);
	return 1;
      } else {
	charMapAdd(cmap, &rcstate->classtab[idx-CS_LAST]);
	return 1;
      }
    }
      
    case STAR:
    case LSTAR: {
      int idx = UCHAR(OPERAND(next)[0]);
      if (idx == UNOPTIMIZED) {
	if (!calcNextEatenByte(rcstate, regoptnext(next), cmap)) {
	  return 0;
	}
	next = OPERAND(next+1);
	break;
      } else if (idx < CS_LAST) {
	charMapAdd(cmap, &bitab[idx]);
	return 1;
      } else {
	charMapAdd(cmap, &rcstate->classtab[idx-CS_LAST]);
	return 1;
      }
    }

    case ANYOF:
      charMapAdd(cmap, &rcstate->classtab[*OPERAND(next)]);
      return 1;

    case BIANYOF:
      charMapAdd(cmap, &bitab[*OPERAND(next)]);
      return 1;

    case LOOK:
    case PLUS:
    case LPLUS:
      /* We will have at least one of OPERAND so point next to it */
      next = OPERAND(next);
      break;

    case REPEAT:
    case LREPEAT: {
      int idx = UCHAR(OPERAND(next)[2]);
      if (idx == UNOPTIMIZED) {
	if (OPERAND(next)[0] == 0) {
	  /* possible 0 repeats */
	  if (!calcNextEatenByte(rcstate, regoptnext(next), cmap)) {
	    return 0;
	  }
	}
	next = OPERAND(next+3);
	break;
      } else if (idx < CS_LAST) {
	charMapAdd(cmap, &bitab[idx]);
	return 1;
      } else {
	charMapAdd(cmap, &rcstate->classtab[idx-CS_LAST]);
	return 1;
      }
    }

    case OREND:
    case ORBRANCH:
      /* Need to check both branches */
      if (!calcNextEatenByte(rcstate, regoptnext(next), cmap)) {
	return 0;
      }
      next = OPERAND(next);
      break;

    case BACK:
      /* Should not have infinite loop because the regexp that
       * is involved with a looping construct must have width
       */
      next = regback(next);
      break;
    case WBOUND:
    case NWBOUND:
    case BOL:
    case BOINPUT:
    case BOS:
    case CLOSE:
    case FASTLOOK:
    case ENDLOOK:
    case NOTHING:
    case OPEN:
      /* try the next one */
      next = regoptnext(next);
      break;

    case ANY:
      /* Could be anything so fail since anything is worst case */
    case EOS:
    case EOL:
      /* EOS and EOL will match the end of input which does not eat a byte */
    case BACKREF:
    case END:
    case NLOOK:
    case FASTNLOOK:
    default:
      /* Don't know what could be next */
      return 0;
    }
  }
}

static int fcdSuccessCount = 0;
static int fcdFailCount = 0;

static void findCharMap(rcstate, modptr, csp)
struct regcomp_state *rcstate;
char* modptr;
struct re_classset *csp;
{
  int i;
  int max;

  /* examine csp and figure out what to do */
  if (csp->cscount >= (CHARMAP_SIZE / 2)) {
    /* probably a negated */
    max = CHARMAP_SIZE - 2;
  } else {
    max = (CHARMAP_SIZE / 2) - 2;
  }
  /* max is now set to the maximum number of chars.
   * If csp contains too many chars then we don't want
   * to bother constructing a lookahead since it will not
   * filter anything out.
   */
  if (csp->cscount < max) {
    fcdSuccessCount++;
    for (i=0; i < CS_LAST; i++) {
      if (csp->cscount == bitab[i].cscount &&
	  csp->cstotal == bitab[i].cstotal &&
	  memcmp(csp->csset, bitab[i].csset, sizeof(csp->csset)) == 0) {
	/* found a builtin that matches */
	modptr[0] = i;
	return;
      }
    }
    for (i=0; i < rcstate->regnClassCount; i++) {
      if (csp->cscount == rcstate->classtab[i].cscount &&
	  csp->cstotal == rcstate->classtab[i].cstotal &&
	  memcmp(csp->csset, rcstate->classtab[i].csset,
		 sizeof(csp->csset)) == 0) {
	/* found one that matches */
	modptr[0] = i + CS_LAST;
	return;
      }
    }
    rcstate->regnClassCount++;
    modptr[0] = i + CS_LAST;
  }
}

static void fcdStartOrCheck(rcstate, orstart)
struct regcomp_state *rcstate;
char* orstart;
{
  struct re_classset *csp;
  char *scan = orstart;
  char *modptr = OPERAND(scan);
  char *operand = OPERAND(scan+1);

  if (UCHAR(*modptr) != UNOPTIMIZED) {
    return; /* already optimized */
  }
  csp = &rcstate->classtab[rcstate->regnClassCount];
  initcs(csp, 0);

  do {
    if (!calcNextEatenByte(rcstate, operand, csp)) {
      fcdFailCount++;
      return;
    }
    scan = regoptnext(scan);
    operand = OPERAND(scan);
  } while (OP(scan) != OREND);
  /* operand now points to the very last alternative */
  if (!calcNextEatenByte(rcstate, operand, csp)) {
    fcdFailCount++;
    return;
  }

  findCharMap(rcstate, modptr, csp);
}

static void fcdBranchCheck(rcstate, modptr, b1, b2)
struct regcomp_state *rcstate;
char* modptr;
char* b1;
char* b2;
{
  struct re_classset *csp;

  if (UCHAR(*modptr) != UNOPTIMIZED) {
    return; /* already optimized */
  }

  csp = &rcstate->classtab[rcstate->regnClassCount];
  initcs(csp, 0);

  if (!calcNextEatenByte(rcstate, b1, csp)) {
    fcdFailCount++;
    return;
  }
  if (!calcNextEatenByte(rcstate, b2, csp)) {
    fcdFailCount++;
    return;
  }

  findCharMap(rcstate, modptr, csp);
}

/* A big savings comes from inserting look ahead right in front
 * of branches if we can figure out the set of all possible characters.
 * This optimization is called 'first character discrimination' FCD.
 * Returns 0 if no more optimization is possible.
 */
static int optimizeNodes(rcstate, scan, stopNode)
struct regcomp_state *rcstate;
char* scan;
char* stopNode;
{
  if (rcstate->regnClassCount >= MAX_CLASS_COUNT) {
    return 0;
  }
  for (;;) {
    if (scan == stopNode) {
      return 1;
    }
    switch (OP(scan)) {
    case BACK:
    case ORBRANCH:
    case OREND:
    case END:
      return 1;

    case ORSTART: {
      char *orstart = scan;
      char *operand = OPERAND(scan+1);
      char *orend = orstart;
      do {
	orend = regoptnext(orend);
      } while (OP(orend) != OREND);
      orend = regoptnext(orend);
      do {
	if (!optimizeNodes(rcstate, operand, orend)) {
	  return 0;
	}
	scan = regoptnext(scan);
	operand = OPERAND(scan);
      } while (OP(scan) != OREND);
      /* operand now points to the very last alternative */
      if (!optimizeNodes(rcstate, operand, stopNode)) {
	return 0;
      }
      fcdStartOrCheck(rcstate, orstart);
      return (rcstate->regnClassCount < MAX_CLASS_COUNT);
    }

    case BRANCH:
    case LBRANCH:
      if (!optimizeNodes(rcstate, OPERAND(scan)+1, stopNode)) {
	return 0;
      }
      if (!optimizeNodes(rcstate, regoptnext(scan), stopNode)) {
	return 0;
      }
      fcdBranchCheck(rcstate, OPERAND(scan),
		     OPERAND(scan)+1, regoptnext(scan));
      return (rcstate->regnClassCount < MAX_CLASS_COUNT);

    case STAR:
    case LSTAR:
      if (!optimizeNodes(rcstate, regoptnext(scan), stopNode)) {
	return 0;
      }
      fcdBranchCheck(rcstate, OPERAND(scan),
		     OPERAND(scan)+1, regoptnext(scan));
      return (rcstate->regnClassCount < MAX_CLASS_COUNT);
      
    case REPEAT:
    case LREPEAT:
      if (!optimizeNodes(rcstate, regoptnext(scan), stopNode)) {
	return 0;
      }
      fcdBranchCheck(rcstate, OPERAND(scan)+2,
		     OPERAND(scan)+3, regoptnext(scan));
      return (rcstate->regnClassCount < MAX_CLASS_COUNT);
      
    case NOTHING:
    default:
      scan = regoptnext(scan);
      break;
    }
  }
}

static int implicitLineAnchorCheck(scan)
char* scan;
{
  char* candidate = NULL;
  while (scan != NULL && candidate == NULL) {
    switch(OP(scan)) {
    case PLUS:
    case LPLUS:
      candidate = OPERAND(scan);
      break;
    case STAR:
    case LSTAR:
      candidate = OPERAND(scan)+1;
      break;
    case REPEAT:
    case LREPEAT:
      if (UCHAR(OPERAND(scan)[1]) == 0) {
	candidate = OPERAND(scan+3);
      } else {
	return 0;
      }
      break;
    case NOTHING:
    case OPEN:
    case CLOSE:
      /* ignore it and check next node */
      break;
    default:
      return 0; /* no implicit line anchor */
    }
    scan = regnext(scan);
  }
  if (candidate) {
    if (OP(candidate) == ANY) {
      return 1;
    }
  }
  return 0;
}

static void optimize(rcstate, scan)
struct regcomp_state *rcstate;
char* scan;
{
  if (implicitLineAnchorCheck(scan)) {
    if (reginsert(BOINPUT, scan, rcstate, 3)) {
      regtail(scan, scan+3);
    }
  }

  if (!(rcstate->extflags & REGEXP_QUICKCOMPILE)) {
    do {
      fcdSuccessCount = 0;
      fcdFailCount = 0;
    } while (optimizeNodes(rcstate, scan, NULL) &&
	     fcdSuccessCount > 0 && fcdFailCount > 0);
  }
}


/*
 * TclNewRegExec and friends
 */

/*
 * Global work variables for TclNewRegExec().
 */
struct regexec_state  {
    char *reginput;		/* String-input pointer. */
    char *regbol;		/* Beginning of string, for ^, \A check. */
    char *reginstart;		/* Beginning of input, for \G check. */
    char *reginend;             /* Last character of input. */
    char *classptr;
    char *regprog;
    int subexpCount;
    int nocase;
#ifdef DEBUG
    int regflags;
#endif
    Tcl_RegExpExecResult *resptr;
    int *couldHavePtr;
};

/*
 * Forwards.
 */
static int 		regtry _ANSI_ARGS_((char *string,
			    struct regexec_state *restate));
static int 		regmatch _ANSI_ARGS_((char *scan,
			    struct regexec_state *restate));
static char* 		regrepeat _ANSI_ARGS_((char *p, char *scan,
			    struct regexec_state *restate));
static int  repeatmatch _ANSI_ARGS_((struct regexec_state *restate,
				     char *repeatnode, char *next,
				     int min, int max, int isGreedy));
static int lookmatch _ANSI_ARGS_((char* scan,
				  CONST struct regexec_state *restate));
static int couldNotMatchAnyOf _ANSI_ARGS_((struct regexec_state *restate,
					   char* n, int idx));
static int couldNotMatchBI _ANSI_ARGS_((struct regexec_state *restate,
					char* n, int bicode));
static int couldNotMatchChar _ANSI_ARGS_((struct regexec_state *restate,
					  char* n, int c));
static void printRangeChar _ANSI_ARGS_((char* buff, int c));
static void dumpRange _ANSI_ARGS_((char* buff, CONST char* cmap));

/* Case insensitive version of memchr. 'c' is know to be lowercase.
 * endPtr points to the character after the last one in the s1 string.
 */
static char*
memcasechr(s1, c, endPtr)
char* s1;
int c;
char* endPtr;
{
  while (s1 < endPtr) {
    if (TOLC(*s1) == c) {
      return s1;
    }
    s1++;
  }
  return NULL;
}

/* memcasematch returns 1 if the two strings match.
 * Otherwise 0 is returned.
 * 'lcs' has already been convert to lower case.
 * A case-insensitive comparison is done.
 * At most 'n' characters are compared.
 */
static int
memcasematch(s, lcs, n)
CONST char* s;
CONST char* lcs;
int n;
{
  while (n > 0) {
    int c;
    c = TOLC(*s++);
    if (c != *lcs++) {
      return 0; /* does not match */
    }
    n--;
  }
  /* The first n characters matched so we have a match */
  return 1;
}

static char* anyofScan(s, anyofPtr, endPtr)
char *s;
CONST char* anyofPtr;
char* endPtr;
{
  while (s < endPtr) {
    if (anyofPtr[UCHAR(*s)]) {
      return s;
    }
    s++;
  }
  return NULL;
}

static char* isNotCharScan(s, c, endPtr)
char *s;
int c;
char* endPtr;
{
  while (s < endPtr) {
    if (UCHAR(*s) != c) {
      return s;
    }
    s++;
  }
  return NULL;
}

static char* sowScan(s, endPtr)
char *s;
char* endPtr;
{
  while (s < endPtr && BITEST(CS_NOTWORD, *s)) {
    s++;
  }
  if (s >= endPtr) {
    return NULL;
  } else {
    return s;
  }
}

static char* wboundScan(s, endPtr, startPtr)
char *s;
char* endPtr;
char* startPtr;
{
  if (s >= endPtr) {
    return NULL;
  } else if (s == startPtr || BITEST(CS_NOTWORD, s[-1])) {
    /* eat as many nonword chars as possible */
    while (s < endPtr && BITEST(CS_NOTWORD, *s)) {
      s++;
    }
    if (s >= endPtr) {
      return NULL;
    }
  } else {
    /* eat as many word chars as possible */
    while (s < endPtr && BITEST(CS_WORD, *s)) {
      s++;
    }
    /* Note that s == endPtr in this case is a word boundary */
  }
  return s;
}

/*
 - TclNewRegExec - match a regexp against a string
 */
int
TclNewRegExec(prog, string, start, resptr)
register regexp *prog;
register char *string;
char *start;
Tcl_RegExpExecResult* resptr;
{
  return TclNewRegExecSz(prog, string, start, -1, resptr, NULL);
}

/* Don't scan forward since we could match
 * if more input was available.
 */
#define COULD_HAVE_MATCHED_CHECK(p)  if (p && p[0]) {return (0);}
#define COULD_HAVE_MATCHED_CHECKEX(p, v)  \
  if (p && p[0]) {p[0] = (int)v; return (0);}
#define SET_COULD_HAVE_MATCHED(p)  if (p) {p[0] = 1;}
#define DO_COULD_HAVE_CHECK(p) \
  if (p->couldHavePtr && (p->reginput != p->reginend)) {p->couldHavePtr[0] = 1;}

int
TclNewRegExecSz(prog, string, start, length, resptr, couldHavePtr)
register regexp *prog;
register char *string;
char *start;
int length;
Tcl_RegExpExecResult* resptr;
int *couldHavePtr;
{
	char *s;
	struct regexec_state state;
	struct regexec_state *restate= &state;
	char *stringEnd;
	char *anyofPtr;

	if (couldHavePtr) {
	  *couldHavePtr = 0;
	}
	/* Be paranoid... */
	if (prog == NULL || string == NULL) {
		TclNewRegError("NULL parameter");
		return(0);
	}

	if (length == -1) {
	  length = strlen(string);
	}
	stringEnd = string + length;

	state.classptr = prog->regclasses;
	/* Check validity of program. */
	if (UCHAR(prog->program[0]) != MAGIC) {
		TclNewRegError("corrupted program");
		return(0);
	}

	state.resptr = resptr;

	/* If there is a "must appear" string, look for it. */
	if (prog->regmust != NULL && couldHavePtr == NULL) {
	  int fc = prog->regmust[0];
	  size_t cmplen = (size_t)prog->regmlen;
	  s = string;
	  if (prog->regflags & REGEXP_NOCASE) {
	    if (isalpha(fc)) {
	      while ((s = memcasechr(s, fc, stringEnd)) != NULL) {
		if (cmplen == 1) {
		  break;	/* Found it. */
		}
		if ((int)cmplen > (stringEnd - s)) {
		  /* could not match so quit */
		  s = NULL;
		  break;
		}
		if (memcasematch(s, prog->regmust, cmplen)) {
		  break;	/* Found it. */
		}
		s++;
	      }
	    } else {
	      while ((s = (char*)memchr(s, fc, stringEnd - s)) != NULL) {
		if (cmplen == 1) {
		  break;	/* Found it. */
		}
		if ((int)cmplen > (stringEnd - s)) {
		  /* could not match so quit */
		  s = NULL;
		  break;
		}
		if (memcasematch(s, prog->regmust, cmplen)) {
		  break;	/* Found it. */
		}
		s++;
	      }
	    }
	  } else {
	    while ((s = (char*)memchr(s, fc, stringEnd - s)) != NULL) {
	      if (cmplen == 1) {
		break;	/* Found it. */
	      }
	      if ((int)cmplen > (stringEnd - s)) {
		/* could not match so quit */
		s = NULL;
		break;
	      }
	      if (memcmp(s, prog->regmust, cmplen) == 0) {
		break;	/* Found it. */
	      }
	      s++;
	    }
	  }
	  if (s == NULL) {
	    /* Not present. */
	    return(0);
	  }
	}

	/* Mark beginning of string for ^ . */
	restate->regbol = start;
	/* Mark beginning of input for \G . */
	restate->reginstart = string;
	restate->reginend = stringEnd;
	restate->regprog = prog->program+1;
	restate->subexpCount = prog->subexpCount-1;
	restate->nocase = prog->regflags & REGEXP_NOCASE;
#ifdef DEBUG
	restate->regflags = prog->regflags;
#endif
	restate->couldHavePtr = couldHavePtr;

	s = string;
	if (prog->regstartnode) {
	  int c;
	  switch (OP(prog->regstartnode)) {
	  case EXACTLY:
	    c = OPERAND(prog->regstartnode)[2];
	    while ((s <= stringEnd) &&
		   ((s = (char*)memchr(s, c, stringEnd - s)) != NULL)) {
	      if (regtry(s, restate)) {
		return(1);
	      }
	      COULD_HAVE_MATCHED_CHECKEX(couldHavePtr, s);
	      s++;
	    }
	    return (0);
	  case NCEXACTLY:
	    c = OPERAND(prog->regstartnode)[2];
	    if (isalpha(c)) {
	      while ((s = memcasechr(s, c, stringEnd)) != NULL) {
		if (regtry(s, restate)) {
		  return(1);
		}
		COULD_HAVE_MATCHED_CHECKEX(couldHavePtr, s);
		s++;
	      }
	    } else {
	      while ((s <= stringEnd) &&
		     ((s = (char*)memchr(s, c, stringEnd - s)) != NULL)) {
		if (regtry(s, restate)) {
		  return(1);
		}
		COULD_HAVE_MATCHED_CHECKEX(couldHavePtr, s);
		s++;
	      }
	    }
	    return (0);
	  case BIANYOF:
	    c = *OPERAND(prog->regstartnode);
	    anyofPtr = BIANYOFPTR(c);
	    while ((s = anyofScan(s, anyofPtr, stringEnd)) != NULL) {
	      if (regtry(s, restate)) {
		return(1);
	      }
	      COULD_HAVE_MATCHED_CHECKEX(couldHavePtr, s);
	      s++;
	    }
	    return (0);
	  case ANYOF:
	    anyofPtr = ANYOFPTR(prog->regclasses,
				*OPERAND(prog->regstartnode));
	    while ((s = anyofScan(s, anyofPtr, stringEnd)) != NULL) {
	      if (regtry(s, restate)) {
		return(1);
	      }
	      COULD_HAVE_MATCHED_CHECKEX(couldHavePtr, s);
	      s++;
	    }
	    return (0);
	  case ISNOTCHAR:
	    c = UCHAR(OPERAND(prog->regstartnode)[0]);
	    while ((s = isNotCharScan(s, c, stringEnd)) != NULL) {
	      if (regtry(s, restate)) {
		return(1);
	      }
	      COULD_HAVE_MATCHED_CHECKEX(couldHavePtr, s);
	      s++;
	    }
	    return (0);
	  case BRANCH:
	  case LBRANCH:
	  case ORSTART:
	  case STAR:
	  case LSTAR:
	    c = UCHAR(*OPERAND(prog->regstartnode));
	    if (c < CS_LAST) {
	      anyofPtr = BIANYOFPTR(c);
	    } else {
	      anyofPtr = ANYOFPTR(prog->regclasses, (c-CS_LAST));
	    }
	    while ((s = anyofScan(s, anyofPtr, stringEnd)) != NULL) {
	      if (regtry(s, restate)) {
		return(1);
	      }
	      COULD_HAVE_MATCHED_CHECKEX(couldHavePtr, s);
	      s++;
	    }
	    return (0);
	  case REPEAT:
	  case LREPEAT:
	    c = UCHAR(OPERAND(prog->regstartnode)[2]);
	    if (c < CS_LAST) {
	      anyofPtr = BIANYOFPTR(c);
	    } else {
	      anyofPtr = ANYOFPTR(prog->regclasses, (c-CS_LAST));
	    }
	    while ((s = anyofScan(s, anyofPtr, stringEnd)) != NULL) {
	      if (regtry(s, restate)) {
		return(1);
	      }
	      COULD_HAVE_MATCHED_CHECKEX(couldHavePtr, s);
	      s++;
	    }
	    return (0);
	  case BOS:
	  case BOINPUT:
	    return regtry(s, restate);
	  case BOL:
	    if (regtry(s, restate)) {
	      return (1);
	    }
	    COULD_HAVE_MATCHED_CHECKEX(couldHavePtr, s);
	    /* now scan for end of line */
	    while ((s <= stringEnd) &&
		   (s = (char*)memchr(s, '\n', stringEnd - s)) != NULL) {
	      s++; /* skip the \n */
	      if (regtry(s, restate)) {
		return(1);
	      }
	      COULD_HAVE_MATCHED_CHECKEX(couldHavePtr, s);
	    }
	    return 0;
	  case SOW:
	    /* The first check is special */
	    if (s != stringEnd && BITEST(CS_WORD, *s)) {
	      if (regtry(s, restate)) {
		return(1);
	      }
	      COULD_HAVE_MATCHED_CHECKEX(couldHavePtr, s);
	      s++;
	    }
	    while ((s = sowScan(s, stringEnd)) != NULL) {
	      if (regtry(s, restate)) {
		return(1);
	      }
	      COULD_HAVE_MATCHED_CHECKEX(couldHavePtr, s);
	      s++;
	    }
	    return 0;
	      
	  case WBOUND:
	    /* The first check is special */
	    if (s != stringEnd && BITEST(CS_WORD, *s)) {
	      if (regtry(s, restate)) {
		return(1);
	      }
	      COULD_HAVE_MATCHED_CHECKEX(couldHavePtr, s);
	      s++;
	    }
	    while ((s = wboundScan(s, stringEnd, restate->reginstart)) != NULL) {
	      if (regtry(s, restate)) {
		return(1);
	      }
	      COULD_HAVE_MATCHED_CHECKEX(couldHavePtr, s);
	      s++;
	    }
	    return 0;

	  case NWBOUND:
	    /* its not easy to scan for this so it is currenty
	     * not optimized.
	     */
	  case EOS:
	  case EOL:
	  case EOW:
	    /* EOS, EOL, EOW at start is wierd so don't bother optimizing */
	  default:
	    /* use default case below */
	    break;
	  }
	}

	/* Doesn't start with something easy to find -- general case. */
	while (s <= restate->reginend) {
	  if (regtry(s, restate)) {
	    return(1);
	  }
	  COULD_HAVE_MATCHED_CHECKEX(couldHavePtr, s);
	  s++;
	}
	/* Failure. */
	return (0);
}

/*
 - regtry - try match at specific point
 */
static int			/* 0 failure, 1 success */
regtry(string, restate)
char *string;
struct regexec_state *restate;
{
	int i;

	restate->reginput = string;

	i = restate->subexpCount;
	if (i > 0) {
	  char **sp;
	  char **ep;
	  sp = restate->resptr->startp+1;
	  ep = restate->resptr->endp+1;
	  do {
	  *sp++ = NULL;
	  *ep++ = NULL;
	  } while (--i > 0);
	}
	if (regmatch(restate->regprog, restate)) {
		restate->resptr->startp[0] = string;
		restate->resptr->endp[0] = restate->reginput;
		if (restate->couldHavePtr) {
		  restate->couldHavePtr[0] = 0;
		}
		return(1);
	} else {
		return(0);
	}
}

static int lookmatch(scan, restate)
char* scan;
CONST struct regexec_state *restate;
{
  /* Doing lookahead so don't change anything in restate */
  char* inptr = restate->reginput;
  if (inptr == restate->reginend) {
    SET_COULD_HAVE_MATCHED(restate->couldHavePtr);
    return 0;
  }
  switch (OP(scan)) {
  case ANY:
    return 1;
  case EXACTLY: {
    register int len;
    register char *opnd;

    opnd = OPERAND(scan);
    /* Inline the first character, for speed. */
    if (opnd[2] != *inptr) {
      return 0;
    }
    len = GET_EXACT_LEN(opnd);
    if (len > 1) {
      if (len > (restate->reginend - inptr)) {
	return 0;
      }
      if (memcmp(opnd+2, inptr, (size_t) len) != 0) {
	return 0;
      }
    }
    return 1;
  }
  case NCEXACTLY: {
    register int len;
    register char *opnd;

    opnd = OPERAND(scan);
    /* Inline the first character, for speed. */
    if (opnd[2] != TOLC(*inptr)) {
      return 0;
    }
    len = GET_EXACT_LEN(opnd);
    if (len > 1) {
      if (len > (restate->reginend - inptr)) {
	return 0;
      }
      if (!memcasematch(inptr, opnd+2, len)) {
	return 0;
      }
    }
    return 1;
  }
  case ISNOTCHAR:
    return (UCHAR(OPERAND(scan)[0]) != UCHAR(inptr[0]));
  case ANYOF:
    return (ANYOFTEST(restate->classptr, *OPERAND(scan), inptr[0]));
  case BIANYOF:
    return (BITEST(*OPERAND(scan), inptr[0]));
  default:
    fprintf(stderr, "unexpected node %d in lookmatch\n", OP(scan));
    TclNewRegError("unexpected node in lookmatch");
    return 0;
  }
}

/*
 - regmatch - main matching routine
 *
 * Conceptually the strategy is simple:  check to see whether the current
 * node matches, call self recursively to see whether the rest matches,
 * and then act accordingly.  In practice we make some effort to avoid
 * recursion, in particular by going through "ordinary" nodes (that don't
 * need to know whether the rest of the match failed) by a loop instead of
 * by recursion.
 */
#ifdef DEBUG
#define RM_RET(v) { \
    int tmp = (v); \
    if (regnarrate) fprintf(stderr, "%s)\n", tmp ?"true":"false"); \
    return (tmp); }
#else
#define RM_RET(v)  return (v)
#endif

static int			/* 0 failure, 1 success */
regmatch(scan, restate)
char *scan;
struct regexec_state *restate;
{
    char *save;
    char *next;		/* Next node. */
    int val;
#ifdef DEBUG
    int regnarrate = restate->regflags & REGEXP_TRACE;
    if (scan != NULL && regnarrate) {
      fprintf(stderr, "(%d\n", restate->reginput - restate->regbol);
    }
#endif
    while (scan != NULL) {
#ifdef DEBUG
        if (regnarrate) {
	  fprintf(stderr, "%s...\n", regprop(scan));
	}
#endif
	switch (OP(scan)) {
	    case BOS:
		if (restate->reginput != restate->regbol) {
		    RM_RET(0);
		}
		scan = regoptnext(scan);
		break;
	    case EOS:
		if (restate->reginput != restate->reginend) {
		    RM_RET(0);
		}
		scan = regoptnext(scan);
		break;
	    case BOL:
		if (restate->reginput != restate->regbol &&
		    restate->reginput[-1] != '\n') {
		  DO_COULD_HAVE_CHECK(restate);
		  RM_RET(0);
		}
		scan = regoptnext(scan);
		break;
	    case EOL:
		if (restate->reginput != restate->reginend &&
		    *restate->reginput != '\n') {
		  DO_COULD_HAVE_CHECK(restate);
		  RM_RET(0);
		}
		scan = regoptnext(scan);
		break;
	    case BOINPUT:
		if (restate->reginput != restate->reginstart) {
		    RM_RET(0);
		}
		scan = regoptnext(scan);
		break;
	    case WBOUND:
	      if (restate->reginput == restate->reginend ||
		  !BITEST(CS_WORD, restate->reginput[0])) {
		/* not a word to the right so check for word to left */
		if (restate->reginput != restate->regbol &&
		    BITEST(CS_WORD, restate->reginput[-1])) {
		  /* we are at the end of a word */
		  scan = regoptnext(scan);
		  break;
		}
	      } else {
		/* a word to the right so check for not a word to left */
		if (restate->reginput == restate->regbol ||
		    !BITEST(CS_WORD, restate->reginput[-1])) {
		  /* we are at the start of a word */
		  scan = regoptnext(scan);
		  break;
		}
	      }
	      /* not at the start or end of a word so fail. */
	      DO_COULD_HAVE_CHECK(restate);
	      RM_RET(0);
	    case NWBOUND:
	      if (restate->reginput == restate->reginend ||
		  !BITEST(CS_WORD, restate->reginput[0])) {
		/* not a word to the right so check for word to left */
		if (restate->reginput != restate->regbol &&
		    BITEST(CS_WORD, restate->reginput[-1])) {
		  /* we are at the end of a word so fail*/
		  DO_COULD_HAVE_CHECK(restate);
		  RM_RET(0);
		}
	      } else {
		/* a word to the right so check for not a word to left */
		if (restate->reginput == restate->regbol ||
		    !BITEST(CS_WORD, restate->reginput[-1])) {
		  /* we are at the start of a word so fail*/
		  RM_RET(0);
		}
	      }
	      /* not at the start or end of a word so keep going */
	      scan = regoptnext(scan);
	      break;
	    case SOW:
		if ((restate->reginput == restate->regbol ||
		     !BITEST(CS_WORD, restate->reginput[-1])) &&
		    (restate->reginput != restate->reginend &&
		     BITEST(CS_WORD, restate->reginput[0]))) {
		  /* we are at the start of a word */
		  scan = regoptnext(scan);
		  break;
		} else {
		  DO_COULD_HAVE_CHECK(restate);
		  RM_RET(0);
		}
	    case EOW:
		if ((restate->reginput == restate->reginend ||
		     !BITEST(CS_WORD, restate->reginput[0])) &&
		    restate->reginput != restate->regbol &&
		    BITEST(CS_WORD, restate->reginput[-1])) {
		  /* we are at the end of a word */
		  scan = regoptnext(scan);
		  break;
		} else {
		  DO_COULD_HAVE_CHECK(restate);
		  RM_RET(0);
		}
	    case ANY:
		if (restate->reginput == restate->reginend) {
		  SET_COULD_HAVE_MATCHED(restate->couldHavePtr);
		  RM_RET(0);
		}
		restate->reginput++;
		scan = regoptnext(scan);
		break;
	    case EXACTLY: {
		register int len;
		register char *opnd;

		if (restate->reginput == restate->reginend) {
		  SET_COULD_HAVE_MATCHED(restate->couldHavePtr);
		  RM_RET(0);
		}
		opnd = OPERAND(scan);
		/* Inline the first character, for speed. */
		if (opnd[2] != *restate->reginput) {
		  RM_RET(0);
		}
		len = GET_EXACT_LEN(opnd);
		if (len > 1) {
		  int remlen = restate->reginend - restate->reginput;
		  if (len > remlen) {
		    /* see if all the remaining bytes match */
		    if (restate->couldHavePtr &&
			memcmp(opnd+2, restate->reginput, (size_t)remlen) == 0) {
		      *restate->couldHavePtr = 1;
		    }
		    RM_RET(0);
		  }
		  if (memcmp(opnd+2, restate->reginput, (size_t) len) != 0) {
		    RM_RET(0);
		  }
		}
		restate->reginput += len;
		scan = regoptnext(scan);
		break;
	    }
	    case NCEXACTLY: {
		register int len;
		register char *opnd;

		if (restate->reginput == restate->reginend) {
		  SET_COULD_HAVE_MATCHED(restate->couldHavePtr);
		  RM_RET(0);
		}
		opnd = OPERAND(scan);
		/* Inline the first character, for speed. */
		if (opnd[2] != TOLC(*restate->reginput)) {
		  RM_RET(0);
		}
		len = GET_EXACT_LEN(opnd);
		if (len > 1) {
		  int remlen = restate->reginend - restate->reginput;
		  if (len > remlen) {
		    /* see if all the remaining bytes match */
		    if (restate->couldHavePtr &&
			memcasematch(restate->reginput, opnd+2, remlen)) {
		      *restate->couldHavePtr = 1;
		    }
		    RM_RET(0);
		  }
		  if (!memcasematch(restate->reginput, opnd+2, len)) {
		    RM_RET(0);
		  }
		}
		restate->reginput += len;
		scan = regoptnext(scan);
		break;
	    }
	    case ISNOTCHAR:
		if (restate->reginput == restate->reginend ||
		    UCHAR(OPERAND(scan)[0]) == UCHAR(*restate->reginput)) {
		    RM_RET(0);
		}
		restate->reginput++;
		scan = regoptnext(scan);
		break;
	    case ANYOF:
	        if (restate->reginput == restate->reginend) {
		  SET_COULD_HAVE_MATCHED(restate->couldHavePtr);
		  RM_RET(0);
	        }
		if(!ANYOFTEST(restate->classptr, *OPERAND(scan),
			      *restate->reginput)) {
		  RM_RET(0);
		}
		restate->reginput++;
		scan = regoptnext(scan);
		break;
	    case BIANYOF:
	        if (restate->reginput == restate->reginend) {
		  SET_COULD_HAVE_MATCHED(restate->couldHavePtr);
		  RM_RET(0);
	        }
		if (!BITEST(*OPERAND(scan), *restate->reginput)) {
		  RM_RET(0);
		}
		restate->reginput++;
		scan = regoptnext(scan);
		break;
	    case NOTHING:
		scan = regoptnext(scan);
		break;
	    case BACK:
		scan = regback(scan);
		break;
            case LOOK:
		save = restate->reginput;
		if (!regmatch(OPERAND(scan),restate)) {
		  RM_RET(0);
		}
		restate->reginput = save;
		scan = regoptnext(scan);
		scan = regoptnext(scan); /* skip ENDLOOK */
		break;
            case NLOOK:
		save = restate->reginput;
		if (regmatch(OPERAND(scan),restate)) {
		  RM_RET(0);
		}
		/* Make sure we are sure the negated match could not match */
		COULD_HAVE_MATCHED_CHECK(restate->couldHavePtr);
		scan = regoptnext(scan);
		scan = regoptnext(scan); /* skip ENDLOOK */
		break;
            case FASTLOOK:
		scan = regoptnext(scan);
		if (!lookmatch(scan, restate)) {
		  RM_RET(0);
		}
		scan = regoptnext(scan);
		break;
            case FASTNLOOK:
		scan = regoptnext(scan);
		if (lookmatch(scan, restate)) {
		  RM_RET(0);
		}
		/* Make sure we are sure the negated match could not match */
		COULD_HAVE_MATCHED_CHECK(restate->couldHavePtr);
		scan = regoptnext(scan);
		break;
	    case ENDLOOK:
	      /* If we got this far then let LOOK know about success */
	      RM_RET(1);
	    case BACKREF: {
	        CONST char *startp;
	        CONST char *endp;
	        CONST char *s;
		int len;
		int remlen;
		val = UCHAR(*OPERAND(scan));
		startp = restate->resptr->startp[val];
		endp = restate->resptr->endp[val];
		s = restate->reginput;
		if (startp == NULL || endp == NULL) {
		  /* Fail since referenced expression did not match */
		  RM_RET(0);
		}
		len = endp - startp;
		if (len < 0) {
		  RM_RET(0);
		} else if (len == 0) {
		  /* Always match empty text */
		  scan = regoptnext(scan);
		  break;
		}

		remlen = (restate->reginend - s);
		if (len > remlen) {
		  if (restate->couldHavePtr) {
		    /* check to see if remaining chars do match */
		    if (restate->nocase) {
		      while (remlen > 0) {
			if (TOLC(*s++) != TOLC(*startp++)) {
			  RM_RET(0); /* does not match */
			}
			remlen--;
		      }
		    } else if (memcmp(s, startp, (size_t)remlen) != 0) {
		      RM_RET(0);
		    }
		    *restate->couldHavePtr = 1;
		  }
		  RM_RET(0);
		}
		if (restate->nocase) {
		  do {
		    if (TOLC(*s++) != TOLC(*startp++)) {
		      RM_RET(0); /* does not match */
		    }
		  } while (startp != endp);
		} else if (memcmp(s, startp, (size_t)len) != 0) {
		  RM_RET(0);
		}
		restate->reginput += len;
		scan = regoptnext(scan);
		break;
	    }
	    case OPEN:
		val = UCHAR(*OPERAND(scan));
		save = restate->resptr->startp[val];
		/* Tentative but set it now for backreferences */
		restate->resptr->startp[val] = restate->reginput;

		scan = regoptnext(scan);
		if (regmatch(scan,restate)) {
		  RM_RET(1);
		} else {
		  restate->resptr->startp[val] = save;
		  RM_RET(0);
		}
	    case CLOSE:
		val = UCHAR(*OPERAND(scan));
		save = restate->resptr->endp[val];
		/* Tentative but set it now for backreferences */
		restate->resptr->endp[val] = restate->reginput;

		scan = regoptnext(scan);
		if (regmatch(scan,restate)) {
		  RM_RET(1);
		} else {
		  restate->resptr->endp[val] = save;
		  RM_RET(0);
		}
	    case ORSTART: {
	      char * operand = OPERAND(scan);
	      val = UCHAR(*operand);
	      if (val != UNOPTIMIZED) {
		if (restate->reginput == restate->reginend) {
		  SET_COULD_HAVE_MATCHED(restate->couldHavePtr);
		  RM_RET(0);
		}
		if (val < CS_LAST) {
		  if (!BITEST(val, *restate->reginput)) {
		    RM_RET(0);
		  }
		} else {
		  if (!ANYOFTEST(restate->classptr, (val-CS_LAST),
				 *restate->reginput)) {
		    RM_RET(0);
		  }
		}
	      }
	      operand++;
	      save = restate->reginput;
	      do {
		if (regmatch(operand, restate)) {
		  RM_RET(1);
		}
		/* If the prev regmatch could have match if it had
		 * more input then fail now so top level can get more input.
		 */
		COULD_HAVE_MATCHED_CHECK(restate->couldHavePtr);
		restate->reginput = save;
		scan = regoptnext(scan);
		operand = OPERAND(scan);
	      } while (OP(scan) != OREND);
	      /* operand now points to the very last alternative */
	      scan = operand;
	      break;
	    }

#define LAZY 0
#define GREEDY 1
	    case PLUS:
	      next = regoptnext(scan);
	      RM_RET(repeatmatch(restate, OPERAND(scan), next,
				 1, -1, GREEDY));
	    case LPLUS:
	      next = regoptnext(scan);
	      RM_RET(repeatmatch(restate, OPERAND(scan), next,
				 1, -1, LAZY));
	    case STAR:
 	    case LSTAR:
	      val = UCHAR(OPERAND(scan)[0]);
	      if (val != UNOPTIMIZED) {
		if (restate->reginput == restate->reginend) {
		  SET_COULD_HAVE_MATCHED(restate->couldHavePtr);
		  RM_RET(0);
		}
		if (val < CS_LAST) {
		  if (!BITEST(val, *restate->reginput)) {
		    RM_RET(0);
		  }
		} else {
		  if (!ANYOFTEST(restate->classptr, (val-CS_LAST),
				 *restate->reginput)) {
		    RM_RET(0);
		  }
		}
	      }
	      next = regoptnext(scan);
	      RM_RET(repeatmatch(restate, OPERAND(scan)+1, next,
				 0, -1,
				 (OP(scan)==STAR) ? GREEDY : LAZY));
	    case REPEAT:
	    case LREPEAT:
	      val = UCHAR(OPERAND(scan)[2]);
	      if (val != UNOPTIMIZED) {
		if (restate->reginput == restate->reginend) {
		  SET_COULD_HAVE_MATCHED(restate->couldHavePtr);
		  RM_RET(0);
		}
		if (val < CS_LAST) {
		  if (!BITEST(val, *restate->reginput)) {
		    RM_RET(0);
		  }
		} else {
		  if (!ANYOFTEST(restate->classptr, (val-CS_LAST),
				 *restate->reginput)) {
		    RM_RET(0);
		  }
		}
	      }
	      val = UCHAR(OPERAND(scan)[1]);
	      if (val == 0) {
		val = -1;
	      }
	      next = regoptnext(scan);
	      RM_RET(repeatmatch(restate, OPERAND(scan)+3, next,
				 UCHAR(OPERAND(scan)[0]), val,
				 (OP(scan)==REPEAT) ? GREEDY : LAZY));

	    case LBRANCH: {
	      char * operand = OPERAND(scan);
	      val = UCHAR(*operand);
	      if (val != UNOPTIMIZED) {
		if (restate->reginput == restate->reginend) {
		  SET_COULD_HAVE_MATCHED(restate->couldHavePtr);
		  RM_RET(0);
		}
		if (val < CS_LAST) {
		  if (!BITEST(val, *restate->reginput)) {
		    RM_RET(0);
		  }
		} else {
		  if (!ANYOFTEST(restate->classptr, (val-CS_LAST),
				 *restate->reginput)) {
		    RM_RET(0);
		  }
		}
	      }
	      operand++;
	      save = restate->reginput;
	      /* First try the lazy way */
	      if (regmatch(regoptnext(scan),restate)) {
		RM_RET(1);
	      }
	      /* If the prev regmatch could have match if it had
	       * more input then fail now so top level can get more input.
	       */
	      COULD_HAVE_MATCHED_CHECK(restate->couldHavePtr);
	      /* That didn't work so backtrack
	       * and try the path with some data */
	      restate->reginput = save;
	      scan = operand;
	      break;
	    }
	    case BRANCH: {
	      char * operand = OPERAND(scan);
	      val = UCHAR(*operand);
	      if (val != UNOPTIMIZED) {
		if (restate->reginput == restate->reginend) {
		  SET_COULD_HAVE_MATCHED(restate->couldHavePtr);
		  RM_RET(0);
		}
		if (val < CS_LAST) {
		  if (!BITEST(val, *restate->reginput)) {
		    RM_RET(0);
		  }
		} else {
		  if (!ANYOFTEST(restate->classptr, (val-CS_LAST),
				 *restate->reginput)) {
		    RM_RET(0);
		  }
		}
	      }
	      operand++;

	      save = restate->reginput;
	      if (regmatch(operand,restate)) {
		RM_RET(1);
	      }
	      /* If the prev regmatch could have match if it had
	       * more input then fail now so top level can get more input.
	       */
	      COULD_HAVE_MATCHED_CHECK(restate->couldHavePtr);
	      /* That didn't work so backtrack
	       * and try the path with no data
	       */
	      restate->reginput = save;
	      scan = regoptnext(scan);
	      break;
	    }
	    case END:
		RM_RET(1);	/* Success! */
	    default:
#ifdef DEBUG
		  fprintf(stderr, "regmatch unexpected op=%d idx=%d\n",
			  OP(scan), scan - restate->regprog);
#endif
		TclNewRegError("memory corruption");
		RM_RET(0);
	}
    }

    /*
     * We get here only if there's trouble -- normally "case END" is
     * the terminating point.
     */
    TclNewRegError("corrupted pointers");
    RM_RET(0);
}

/* couldNotMatchAnyOf returns 1 if the 'n' node could not
 * possibly match the 'idx' char set.
 */
static int
couldNotMatchAnyOf(restate, n, idx)
struct regexec_state *restate;
char* n;
int idx;
{
  char *classptr = &restate->classptr[idx * CHARMAP_SIZE];
  char *p;
  char *endPtr;
  switch (OP(n)) {
  case ANYOF:
    return !doSetsIntersect(classptr,
			    &restate->classptr[*OPERAND(n) * CHARMAP_SIZE]);
  case BIANYOF:
    return !doSetsIntersect(classptr, bitab[*OPERAND(n)].csset);
  case EXACTLY:
  case NCEXACTLY:
    p = OPERAND(n);
    endPtr = p + 2;
    endPtr += GET_EXACT_LEN(p);
    p += 2;
    while (p < endPtr) {
      if (classptr[UCHAR(*p)]) {
	return 0;
      }
      p++;
    }
    return 1;
  case ISNOTCHAR: {
    int c = UCHAR(OPERAND(n)[0]);
    if (!classptr[c]) {
      /* classptr must have at least 1 byte in it
       * and since the byte is not the only one that ISNOTCHAR
       * can not match it must be at least of the others that
       * ISNOTCHAR could match.
       */
      return 0;
    } else {
      p = (char*)memchr(classptr, 1, CHARMAP_SIZE);
      if (p == &classptr[c]) {
	classptr = p+1;
	p = (char*)memchr(classptr, 1, (CHARMAP_SIZE - c) + 1);
      }
      /* if p == NULL then the only byte in classptr that could match
       * is that one that ISNOTCHAR can not match.
       */
      return p == NULL;
    }
  }
  default:
    return 0;
  }
}

/* couldNotMatchBI returns 1 if the 'n' node could not
 * possibly match the 'bicode' builtin char set.
 */
static int couldNotMatchBI(restate, n, bicode)
struct regexec_state *restate;
char* n;
int bicode;
{
  char *p;
  char *endPtr;
  char *anyofPtr;
  switch (OP(n)) {
  case ANYOF:
    return !doSetsIntersect(&restate->classptr[*OPERAND(n) * CHARMAP_SIZE],
			    bitab[bicode].csset);
  case BIANYOF:
    return !biIntersectTable[bicode][*OPERAND(n)];
  case EXACTLY:
  case NCEXACTLY:
    p = OPERAND(n);
    endPtr = p + 2;
    endPtr += GET_EXACT_LEN(p);
    p += 2;
    anyofPtr = BIANYOFPTR(bicode);
    while (p < endPtr) {
      if (anyofPtr[UCHAR(*p)]) {
	return 0;
      }
      p++;
    }
    return 1;
  case ISNOTCHAR:
    /* All builtins match more than one character.
     * Since a ISNOTCHAR is the same as a charset with only
     * one character missing we know that a ISNOTCHAR could
     * always intersect a builtin.
     */
  default:
    return 0;
  }
}

/* couldNotMatchNotChar returns 1 if the 'n' node could not possibly
 * match any character that is not 'c'.
 */
static int couldNotMatchNotChar(restate, n, c)
struct regexec_state *restate;
char* n;
int c;
{
  switch (OP(n)) {
  case ANYOF: {
    char *classptr = ANYOFPTR(restate->classptr, *OPERAND(n));
    if (!classptr[c]) {
      /* classptr must have at least 1 byte in it
       * and since the byte is not the only one that ISNOTCHAR
       * can not match it must be at least of the others that
       * ISNOTCHAR could match.
       */
      return 0;
    } else {
      char* p = (char*)memchr(classptr, 1, CHARMAP_SIZE);
      if (p == &classptr[c]) {
	classptr = p+1;
	p = (char*)memchr(classptr, 1, (CHARMAP_SIZE - c) + 1);
      }
      /* if p == NULL then the only byte in classptr that could match
       * is that one that ISNOTCHAR can not match.
       */
      return p == NULL;
    }
  }
  case NCEXACTLY:
  case EXACTLY: {
    char *p = OPERAND(n);
    size_t len = GET_EXACT_LEN(p);
    char *endp = p + len;
    p += 2;
    while (p != endp) {
      if (UCHAR(*p) != c) {
	return 0;
      }
      p++;
    }
    return 1;
  }
  case ISNOTCHAR:
    /* two sets of 255 chars must intersect and thus could match */
  case BIANYOF:
    /* All builtins match more than one character.
     * Since a ISNOTCHAR is the same as a charset with only
     * one character missing we know that a ISNOTCHAR could
     * always intersect a builtin.
     */
  default:
    return 0;
  }
}

/* couldNotMatchChar returns 1 if the 'n' node could not possibly
 * match that character 'c'.
 */
static int couldNotMatchChar(restate, n, c)
struct regexec_state *restate;
char* n;
int c;
{
  switch (OP(n)) {
  case ANYOF:
    return !ANYOFTEST(restate->classptr, *OPERAND(n), c);
  case BIANYOF:
    return !BITEST(*OPERAND(n), c);
  case NCEXACTLY:
  case EXACTLY: {
    char *p = OPERAND(n);
    size_t len = GET_EXACT_LEN(p);
    p += 2;
    return ((char*)memchr(p, c, len) == NULL);
  }
  case ISNOTCHAR:
    return c == UCHAR(OPERAND(n)[0]);
  default:
    return 0;
  }
}

/*
 * repeatmatch - used to match STAR, PLUS, LSTAR, and LPLUS.
 * Return 1 on success and 0 on failure.
 */
static int
repeatmatch(restate, repeatnode, next, min, max, isGreedy)
struct regexec_state *restate;
char *repeatnode;
char *next;
int min;
int max;
int isGreedy;
{
  char *startp;
  char *endp;
  char *save;
  int nextop;
  char *nextd;
  char *inend; /* cached restate->reginend */
  char nextch;
  char *opnd;
  int opndval;
  int atend = 0;
  int *couldHavePtr = restate->couldHavePtr;
#ifdef DEBUG
  int regnarrate = 0;
#endif

  save = restate->reginput;
  startp = save + min;
  inend = restate->reginend;
  endp = save;

  opnd = OPERAND(repeatnode);
  switch (OP(repeatnode)) {
  case ANY:
    endp = inend;
    break;
  case EXACTLY:
    opndval = opnd[2];
    while (endp != inend && opndval == *endp) {
      endp++;
    }
    break;
  case NCEXACTLY:
    opndval = opnd[2];
    while (endp != inend && opndval == TOLC(*endp)) {
      endp++;
    }
    break;
  case ANYOF:
    opndval = *opnd;
    opnd = ANYOFPTR(restate->classptr, opndval);
    while (endp != inend && opnd[UCHAR(*endp)]) {
      endp++;
    }
    break;
  case ISNOTCHAR:
    opndval = UCHAR(*opnd);
    while (endp != inend && opndval != UCHAR(*endp)) {
      endp++;
    }
    break;
  case BIANYOF:
    opndval = *opnd;
    opnd = BIANYOFPTR(opndval);
    while (endp != inend && opnd[UCHAR(*endp)]) {
      endp++;
    }
    break;
  default:		/* Oh dear.  Called inappropriately. */
    TclNewRegError("internal foulup");
    break;
  }

  if (endp < startp) {
    /* Didn't find enough to meet 'min' requirement */
    if (couldHavePtr && endp == inend) {
      SET_COULD_HAVE_MATCHED(couldHavePtr);
    }
    return 0;
  }
  if (max != -1) {
    if (endp > (save + max)) {
      endp = save + max;
    }
  }
  if (couldHavePtr) {
    atend = (endp == inend);
    if (atend && isGreedy) {
      /* If more input was available this guy could have matched it.
       * So we need to be pessimistic and fail.
       */
      SET_COULD_HAVE_MATCHED(couldHavePtr);
      return 0;
    }
  }
  /* Now find the next node that will definitely match something */
  nextd = next;
tryNextNode:
  nextop = OP(nextd);
  switch (nextop) {
  case END:
  case BOS:
  case EOS:
  case ANYOF:
  case EXACTLY:
  case NCEXACTLY:
  case BIANYOF:
  case WBOUND:
  case NWBOUND:
  case SOW:
  case EOW:
  case BOL:
  case EOL:
  case BOINPUT:
  case BACKREF:
  case ISNOTCHAR:
    /* We found a definite node. */
    break;

  case LOOK:
  case PLUS:
  case LPLUS:
    /* try the operand */
    nextd = OPERAND(nextd);
    goto tryNextNode;

  case REPEAT:
  case LREPEAT:
    if (UCHAR(OPERAND(nextd)[2]) != UNOPTIMIZED) {
      /* Found a definite node */
      break;
    } else if (OPERAND(nextd)[0] != 0) {
      nextd = OPERAND(nextd)+3;
      goto tryNextNode;
    } else {
      /* next could be more than one thing so return NULL */
      nextd = NULL;
      nextop = NOTHING;
      break;
    }

  case BRANCH:
  case LBRANCH:
  case ORSTART:
  case STAR:
  case LSTAR:
    if (UCHAR(*OPERAND(nextd)) != UNOPTIMIZED) {
      /* Found a definite node */
    } else {
      nextd = NULL;
      nextop = NOTHING;
    }
    break;

  case BACK:
    /* try the next one */
    nextd = regback(nextd);
    goto tryNextNode;

  case NOTHING:
  case OPEN:
  case FASTLOOK:
  case ENDLOOK:
  case CLOSE:
    /* try the next one */
    nextd = regoptnext(nextd);
    goto tryNextNode;

  case ANY:
    /* Its definite but could be anything so...*/
  case NLOOK:
  case FASTNLOOK:
    /* We could scan forward like we do for LOOK
     * but we would need to know to reverse
     * the test of any definite node found. So just give up
     */
  case ORBRANCH:
  case OREND:
    /* nextd could be more than one thing so return NULL */
  default:
    /* Else unknown node */
    /* following regexp nodes too complex for findNextDefiniteNode
     * or the next node is not definite.
     * So set nextop to NOTHING and leave 'next' as is.
     * This tells the code to just do a regmatch.
     */
    nextd = NULL;
    nextop = NOTHING;
    break;
  }
  switch (nextop) {
  case END:
    if (isGreedy) {
      restate->reginput = endp;
    } else { /* lazy */
      restate->reginput = startp;
    }
    if (next == nextd || regmatch(next,restate)) {
      return 1;
    }
    break;
  case BOS:
    if (startp == restate->regbol) {
      restate->reginput = startp;
      if (regmatch(next,restate)) {
	return 1;
      }
    }
    break;
  case BOINPUT:
    if (startp == restate->reginstart) {
      restate->reginput = startp;
      if (regmatch(next,restate)) {
	return 1;
      }
    }
    break;
  case EOS:
    if (endp == inend) {
      restate->reginput = endp;
      if (regmatch(next,restate)) {
	return 1;
      }
    }
    break;
  case BOL:
    if (couldNotMatchChar(restate, repeatnode, '\n')) {
      /* no backtracking required */
#ifdef DEBUG
      if (regnarrate) {
	fprintf(stderr, "  super optimized %s\n", regprop(repeatnode));
      }
#endif
      if (min == 0 && (startp == restate->regbol || startp[-1] == '\n')) {
	restate->reginput = startp;
	if (regmatch(next,restate)) {
	  return 1;
	}
      }
    } else if (isGreedy) {
      if (endp == restate->regbol) {
	restate->reginput = endp;
	if (regmatch(next,restate)) {
	  return 1;
	}
	COULD_HAVE_MATCHED_CHECK(couldHavePtr);
	endp--;
      }
      while (endp >= startp) {
	if (endp[-1] == '\n') {
	  restate->reginput = endp;
	  if (regmatch(next,restate)) {
	    return 1;
	  }
	  COULD_HAVE_MATCHED_CHECK(couldHavePtr);
	}
	/* Couldn't or didn't -- back up. */
	endp--;
      }
      if (startp == restate->regbol) {
	restate->reginput = startp;
	if (regmatch(next,restate)) {
	  return 1;
	}
      }
    } else { /* lazy */
      if (startp == restate->regbol) {
	restate->reginput = startp;
	if (regmatch(next,restate)) {
	  return 1;
	}
	COULD_HAVE_MATCHED_CHECK(couldHavePtr);
	startp++;
      }
      /* no need to consider regbol anymore */
      while (endp >= startp) {
	if (startp[-1] == '\n') {
	  restate->reginput = startp;
	  if (regmatch(next,restate)) {
	    return 1;
	  }
	  COULD_HAVE_MATCHED_CHECK(couldHavePtr);
	}
	/* Couldn't or didn't -- move forward. */
	startp++;
      }
    }
    break;
  case EOL:
    if (couldNotMatchChar(restate, repeatnode, '\n')) {
      /* no backtracking required */
#ifdef DEBUG
      if (regnarrate) {
	fprintf(stderr, "  super optimized %s\n", regprop(repeatnode));
      }
#endif
      isGreedy = 1;
      startp = endp;
    }
    if (isGreedy) {
      if (endp == inend) {
	restate->reginput = endp;
	if (regmatch(next,restate)) {
	  return 1;
	}
	COULD_HAVE_MATCHED_CHECK(couldHavePtr);
	endp--; /* Couldn't or didn't -- back up. */
      }
      while (endp >= startp) {
	if (*endp == '\n') {
	  restate->reginput = endp;
	  if (regmatch(next,restate)) {
	    return 1;
	  }
	  COULD_HAVE_MATCHED_CHECK(couldHavePtr);
	}
	/* Couldn't or didn't -- back up. */
	endp--;
      }
    } else { /* lazy */
      while (endp >= startp) {
	if (startp != inend && *startp == '\n') {
	  restate->reginput = startp;
	  if (regmatch(next,restate)) {
	    return 1;
	  }
	  COULD_HAVE_MATCHED_CHECK(couldHavePtr);
	}
	/* Couldn't or didn't -- move forward. */
	startp++;
      }
      if (endp == inend) {
	restate->reginput = endp;
	if (regmatch(next,restate)) {
	  return 1;
	}
      }
    }
    break;
  case SOW:
    if (endp == inend) {
      endp--;
      if (endp < startp) {
	/* Didn't find enough to meet 'min' requirement */
	SET_COULD_HAVE_MATCHED(couldHavePtr);
	return 0;
      }
    }
    if (couldNotMatchBI(restate, repeatnode, CS_WORD)) {
      /* no backtracking required */
#ifdef DEBUG
      if (regnarrate) {
	fprintf(stderr, "  super optimized %s\n", regprop(repeatnode));
      }
#endif
      isGreedy = 1;
      startp = endp;
    }
    if (isGreedy) {
      while (endp >= startp) {
	if (BITEST(CS_WORD, endp[0]) &&
	    (endp == restate->regbol || !BITEST(CS_WORD, endp[-1]))) {
	  restate->reginput = endp;
	  if (regmatch(next,restate)) {
	    return 1;
	  }
	  COULD_HAVE_MATCHED_CHECK(couldHavePtr);
	}
	/* Couldn't or didn't -- back up. */
	endp--;
      }
    } else { /* lazy */
      if ((startp == restate->regbol) && BITEST(CS_WORD, startp[0])) {
	restate->reginput = startp;
	if (regmatch(next,restate)) {
	  return 1;
	}
	COULD_HAVE_MATCHED_CHECK(couldHavePtr);
	startp++;
      }
      while (endp >= startp) {
	if (!BITEST(CS_WORD, startp[-1]) && BITEST(CS_WORD, startp[0])) {
	  restate->reginput = startp;
	  if (regmatch(next,restate)) {
	    return 1;
	  }
	  COULD_HAVE_MATCHED_CHECK(couldHavePtr);
	}
	/* Couldn't or didn't -- move forward. */
	startp++;
      }
    }
    break;
  case EOW:
    if (couldNotMatchBI(restate, repeatnode, CS_NOTWORD)) {
      /* no backtracking required */
#ifdef DEBUG
      if (regnarrate) {
	fprintf(stderr, "  super optimized %s\n", regprop(repeatnode));
      }
#endif
      isGreedy = 1;
      startp = endp;
    }
    if (isGreedy) {
      while (endp >= startp) {
	if ((endp == inend || !BITEST(CS_WORD, endp[0])) &&
	    endp != restate->regbol && BITEST(CS_WORD, endp[-1])) {
	  /* we are at the end of a word */
	  restate->reginput = endp;
	  if (regmatch(next,restate)) {
	    return 1;
	  }
	  COULD_HAVE_MATCHED_CHECK(couldHavePtr);
	}
	/* Couldn't or didn't -- back up. */
	endp--;
      }
    } else { /* lazy */
      while (endp >= startp) {
	if ((startp == inend || !BITEST(CS_WORD, startp[0])) &&
	    startp != restate->regbol && BITEST(CS_WORD, startp[-1])) {
	  /* we are at the end of a word */
	  restate->reginput = startp;
	  if (regmatch(next,restate)) {
	    return 1;
	  }
	  COULD_HAVE_MATCHED_CHECK(couldHavePtr);
	}
	/* Couldn't or didn't -- move forward. */
	startp++;
      }
    }
    break;
  case WBOUND:
    if (isGreedy) {
      while (endp >= startp) {
	if (endp == inend || !BITEST(CS_WORD, endp[0])) {
	  /* not a word to the right so check for word to left */
	  if (endp != restate->regbol && BITEST(CS_WORD, endp[-1])) {
	    /* we are at the end of a word */
	    restate->reginput = endp;
	    if (regmatch(next,restate)) {
	      return 1;
	    }
	    COULD_HAVE_MATCHED_CHECK(couldHavePtr);
	  }
	} else {
	  /* a word to the right so check for not a word to left */
	  if (endp == restate->regbol || !BITEST(CS_WORD, endp[-1])) {
	    /* we are at the start of a word */
	    restate->reginput = endp;
	    if (regmatch(next,restate)) {
	      return 1;
	    }
	    COULD_HAVE_MATCHED_CHECK(couldHavePtr);
	  }
	}
	/* Couldn't or didn't -- back up. */
	endp--;
      }
    } else { /* lazy */
      while (endp >= startp) {
	if (startp == inend || !BITEST(CS_WORD, startp[0])) {
	  /* not a word to the right so check for word to left */
	  if (startp != restate->regbol && BITEST(CS_WORD, startp[-1])) {
	    /* we are at the end of a word */
	    restate->reginput = startp;
	    if (regmatch(next,restate)) {
	      return 1;
	    }
	    COULD_HAVE_MATCHED_CHECK(couldHavePtr);
	  }
	} else {
	  /* a word to the right so check for not a word to left */
	  if (startp == restate->regbol || !BITEST(CS_WORD, startp[-1])) {
	    /* we are at the start of a word */
	    restate->reginput = startp;
	    if (regmatch(next,restate)) {
	      return 1;
	    }
	    COULD_HAVE_MATCHED_CHECK(couldHavePtr);
	  }
	}
	/* Couldn't or didn't -- move forward. */
	startp++;
      }
    }
    break;
  case NWBOUND:
    if (isGreedy) {
      while (endp >= startp) {
	if (endp == inend || !BITEST(CS_WORD, endp[0])) {
	  /* not a word to the right so check for word to left */
	  if (endp != restate->regbol && BITEST(CS_WORD, endp[-1])) {
	    endp--;
	    continue;
	  }
	} else {
	  /* a word to the right so check for not a word to left */
	  if (endp == restate->regbol || !BITEST(CS_WORD, endp[-1])) {
	    endp--;
	    continue;
	  }
	}
	restate->reginput = endp;
	if (regmatch(next,restate)) {
	  return 1;
	}
	COULD_HAVE_MATCHED_CHECK(couldHavePtr);
	/* Couldn't or didn't -- back up. */
	endp--;
      }
    } else { /* lazy */
      while (endp >= startp) {
	if (startp == inend || !BITEST(CS_WORD, startp[0])) {
	  /* not a word to the right so check for word to left */
	  if (startp != restate->regbol && BITEST(CS_WORD, startp[-1])) {
	    startp++;
	    continue;
	  }
	} else {
	  /* a word to the right so check for not a word to left */
	  if (startp == restate->regbol || !BITEST(CS_WORD, startp[-1])) {
	    startp++;
	    continue;
	  }
	}
	restate->reginput = startp;
	if (regmatch(next,restate)) {
	  return 1;
	}
	COULD_HAVE_MATCHED_CHECK(couldHavePtr);
	/* Couldn't or didn't -- move forward. */
	startp++;
      }
    }
    break;
  case ISNOTCHAR:
    nextch = UCHAR(OPERAND(nextd)[0]);
    if (endp == inend) {
      endp--;
      if (endp < startp) {
	/* Didn't find enough to meet 'min' requirement */
	SET_COULD_HAVE_MATCHED(couldHavePtr);
	return 0;
      }
    }
    if (couldNotMatchNotChar(restate, repeatnode, nextch)) {
      /* no backtracking required */
#ifdef DEBUG
      if (regnarrate) {
	fprintf(stderr, "  super optimized %s\n", regprop(repeatnode));
      }
#endif
      isGreedy = 1;
      startp = endp;
    }
    if (isGreedy) {
      while (endp >= startp) {
	if (nextch != UCHAR(*endp)) {
	  restate->reginput = endp;
	  if (regmatch(next,restate)) {
	    return 1;
	  }
	  COULD_HAVE_MATCHED_CHECK(couldHavePtr);
	}
	/* Couldn't or didn't -- back up. */
	endp--;
      }
    } else { /* lazy */
      while (endp >= startp) {
	if (nextch != UCHAR(*startp)) {
	  restate->reginput = startp;
	  if (regmatch(next,restate)) {
	    return 1;
	  }
	  COULD_HAVE_MATCHED_CHECK(couldHavePtr);
	}
	/* Couldn't or didn't -- move forward. */
	startp++;
      }
    }
    break;
  case ANYOF:
    nextch = *OPERAND(nextd);
  doAnyOf:
    if (endp == inend) {
      endp--;
      if (endp < startp) {
	/* Didn't find enough to meet 'min' requirement */
	SET_COULD_HAVE_MATCHED(couldHavePtr);
	return 0;
      }
    }
    if (couldNotMatchAnyOf(restate, repeatnode, nextch)) {
      /* no backtracking required */
#ifdef DEBUG
      if (regnarrate) {
	fprintf(stderr, "  super optimized %s\n", regprop(repeatnode));
      }
#endif
      isGreedy = 1;
      startp = endp;
    }
    if (isGreedy) {
      char* anyofPtr = ANYOFPTR(restate->classptr, nextch);
      while (endp >= startp) {
	if (anyofPtr[UCHAR(*endp)]) {
	  restate->reginput = endp;
	  if (regmatch(next,restate)) {
	    return 1;
	  }
	  COULD_HAVE_MATCHED_CHECK(couldHavePtr);
	}
	/* Couldn't or didn't -- back up. */
	endp--;
      }
    } else { /* lazy */
      char* anyofPtr = ANYOFPTR(restate->classptr, nextch);
      while (endp >= startp) {
	if (anyofPtr[UCHAR(*startp)]) {
	  restate->reginput = startp;
	  if (regmatch(next,restate)) {
	    return 1;
	  }
	  COULD_HAVE_MATCHED_CHECK(couldHavePtr);
	}
	/* Couldn't or didn't -- move forward. */
	startp++;
      }
    }
    break;
  case BIANYOF:
    nextch = *OPERAND(nextd);
  doBiAnyOf:
    if (endp == inend) {
      endp--;
      if (endp < startp) {
	/* Didn't find enough to meet 'min' requirement */
	SET_COULD_HAVE_MATCHED(couldHavePtr);
	return 0;
      }
    }
    if (couldNotMatchBI(restate, repeatnode, nextch)) {
      /* no backtracking required */
#ifdef DEBUG
      if (regnarrate) {
	fprintf(stderr, "  super optimized %s\n", regprop(repeatnode));
      }
#endif
      isGreedy = 1;
      startp = endp;
    }
    if (isGreedy) {
      char* anyofPtr = BIANYOFPTR(nextch);
      while (endp >= startp) {
	if (anyofPtr[UCHAR(*endp)]) {
	  restate->reginput = endp;
	  if (regmatch(next,restate)) {
	    return 1;
	  }
	  COULD_HAVE_MATCHED_CHECK(couldHavePtr);
	}
	/* Couldn't or didn't -- back up. */
	endp--;
      }
    } else { /* lazy */
      char* anyofPtr = BIANYOFPTR(nextch);
      while (endp >= startp) {
	if (anyofPtr[UCHAR(*startp)]) {
	  restate->reginput = startp;
	  if (regmatch(next,restate)) {
	    return 1;
	  }
	  COULD_HAVE_MATCHED_CHECK(couldHavePtr);
	}
	/* Couldn't or didn't -- move forward. */
	startp++;
      }
    }
    break;
  case BRANCH:
  case LBRANCH:
  case ORSTART:
  case STAR:
  case LSTAR:
    nextch = *OPERAND(nextd);
    if (nextch < CS_LAST) {
      goto doBiAnyOf;
    } else {
      nextch -= CS_LAST;
      goto doAnyOf;
    }
  case REPEAT:
  case LREPEAT:
    nextch = OPERAND(nextd)[2];
    if (nextch < CS_LAST) {
      goto doBiAnyOf;
    } else {
      nextch -= CS_LAST;
      goto doAnyOf;
    }
  case NCEXACTLY:
    nextch = OPERAND(nextd)[2];
  doNcExactly:
    if (endp == inend) {
      endp--;
      if (endp < startp) {
	/* Didn't find enough to meet 'min' requirement */
	SET_COULD_HAVE_MATCHED(couldHavePtr);
	return 0;
      }
    }
    if (couldNotMatchChar(restate, repeatnode, nextch)) {
      /* no backtracking required */
#ifdef DEBUG
      if (regnarrate) {
	fprintf(stderr, "  super optimized %s\n", regprop(repeatnode));
      }
#endif
      isGreedy = 1;
      startp = endp;
    }
    if (isGreedy) {
      while (endp >= startp) {
	if (TOLC(*endp) == nextch) {
	  restate->reginput = endp;
	  if (regmatch(next,restate)) {
	    return 1;
	  }
	  COULD_HAVE_MATCHED_CHECK(couldHavePtr);
	}
	/* Couldn't or didn't -- back up. */
	endp--;
      }
    } else { /* lazy */
      while (endp >= startp) {
	if (TOLC(*startp) == nextch) {
	  restate->reginput = startp;
	  if (regmatch(next,restate)) {
	    return 1;
	  }
	  COULD_HAVE_MATCHED_CHECK(couldHavePtr);
	}
	/* Couldn't or didn't -- move forward. */
	startp++;
      }
    }
    break;
  case EXACTLY:
    nextch = OPERAND(nextd)[2];
  doExactly:
    if (endp == inend) {
      endp--;
      if (endp < startp) {
	/* Didn't find enough to meet 'min' requirement */
	SET_COULD_HAVE_MATCHED(couldHavePtr);
	return 0;
      }
    }
    if (couldNotMatchChar(restate, repeatnode, nextch)) {
      /* no backtracking required */
#ifdef DEBUG
      if (regnarrate) {
	fprintf(stderr, "  super optimized %s\n", regprop(repeatnode));
      }
#endif
      isGreedy = 1;
      startp = endp;
    }
    if (isGreedy) {
      while (endp >= startp) {
	if (*endp == nextch) {
	  restate->reginput = endp;
	  if (regmatch(next,restate)) {
	    return 1;
	  }
	  COULD_HAVE_MATCHED_CHECK(couldHavePtr);
	}
	/* Couldn't or didn't -- back up. */
	endp--;
      }
    } else { /* lazy */
      while (endp >= startp) {
	if (*startp == nextch) {
	  restate->reginput = startp;
	  if (regmatch(next,restate)) {
	    return 1;
	  }
	  COULD_HAVE_MATCHED_CHECK(couldHavePtr);
	}
	/* Couldn't or didn't -- move forward. */
	startp++;
      }
    }
    break;
  case BACKREF: {
    int no = UCHAR(*OPERAND(nextd));
    CONST char *brstartp = restate->resptr->startp[no];
    CONST char *brendp = restate->resptr->endp[no];
    int len;
    if (brstartp != NULL && brendp != NULL) {
      len = brendp - brstartp;
      if (len > 0) {
	if (restate->nocase) {
	  nextch = TOLC(*brstartp);
	  goto doNcExactly;
	} else {
	  nextch = *brstartp;
	  goto doExactly;
	}
      }
    }
    nextop = NOTHING;
    /* FallThrough to default */
  }

  case NOTHING:
  default:
    if (isGreedy) {
      while (endp >= startp) {
	restate->reginput = endp;
	if (regmatch(next,restate)) {
	  return 1;
	}
	COULD_HAVE_MATCHED_CHECK(couldHavePtr);
	/* Couldn't or didn't -- back up. */
	endp--;
      }
    } else { /* lazy */
      while (endp >= startp) {
	restate->reginput = startp;
	if (regmatch(next,restate)) {
	  return 1;
	}
	COULD_HAVE_MATCHED_CHECK(couldHavePtr);
	/* Couldn't or didn't -- move forward. */
	startp++;
      }
    }
    break;
  } /* end switch */
  restate->reginput = save;
  if (atend) {
    *couldHavePtr = 1;
  }
  return 0;
}

/*
 - regrepeat - repeatedly match something simple, report how many
 */
static char*
regrepeat(p, scan, restate)
char *p;
char *scan;
struct regexec_state *restate;
{
	return(scan);
}

/*
 - regnext - dig the "next" pointer out of a node
 */
static char *
regnext(p)
register char *p;
{
  int offset;
  char * result = NULL;

  offset = NEXT(p);
  if (offset != 0) {
    result = p + offset;
    if (OP(p) == BACK) {
      result = p - offset;
    }
  }
  return result;
}

static void printRangeChar(buff, c)
char* buff;
int c;
{
  if (isprint(c)) {
    sprintf(buff+strlen(buff), "%c", c);
  } else if (c == '\n') {
    strcat(buff, "\\n");
  } else if (c == '\t') {
    strcat(buff, "\\t");
  } else {
    sprintf(buff+strlen(buff), "\\x%02x", (unsigned int)c);
  }
}

static void dumpRange(buff, cmap)
char* buff;
CONST char* cmap;
{
  int i;
  int srun = -1;
  strcat(buff, " [");
  if (cmap[CHARMAP_SIZE-1] == 1) {
    /* Its probably negated */
    strcat(buff, "^");
    for (i = 0; i < CHARMAP_SIZE; i++) {
      if (!cmap[i]) {
	if (srun == -1) {
	  srun = i;
	  printRangeChar(buff, i);
	}
      } else {
	if (srun != -1) {
	  if (i != (srun + 1)) {
	    if (i != (srun + 2)) {
	      strcat(buff, "-");
	    }
	    printRangeChar(buff, i-1);
	  }
	  srun = -1;
	}
      }
    }
  } else {
    for (i = 0; i < CHARMAP_SIZE; i++) {
      if (cmap[i]) {
	if (srun == -1) {
	  srun = i;
	  printRangeChar(buff, i);
	}
      } else {
	if (srun != -1) {
	  if (i != (srun + 1)) {
	    if (i != (srun + 2)) {
	      strcat(buff, "-");
	    }
	    printRangeChar(buff, i-1);
	  }
	  srun = -1;
	}
      }
    }
  }
  if (srun != -1 && srun != (CHARMAP_SIZE-1)) {
    if (srun != (CHARMAP_SIZE-2)) {
      strcat(buff, "-");
    }
    printRangeChar(buff, CHARMAP_SIZE-1);
  }
  strcat(buff, "]");
}

/*
 * TclNewRegDump - dump a regexp into a Tcl list which
 * in vaguely comprehensible form
 */
int
TclNewRegDump(interp, r)
Tcl_Interp *interp;
regexp *r;
{
  char buff[256];
  Tcl_Obj *resObjPtr = Tcl_NewObj();
  register char *s;
  register char op = EXACTLY;	/* Arbitrary non-END op. */
  register char *next;

  s = r->program + 1;
  while (op != END) {	/* While that wasn't END last time... */
    op = OP(s);
    strcpy(buff, regprop(r, s)); /* Where, what. */
    next = regnext(s);
    if (next == NULL)		/* Next ptr. */
      strcat(buff, "(0)");
    else 
      sprintf(buff+strlen(buff), "(%d)", (s-r->program)+(next-s));
    s += 3;
    if (op == EXACTLY || op == NCEXACTLY) {
      /* Literal string, where present. */
      char * endPtr = s+2;
      endPtr += GET_EXACT_LEN(s);
      s += 2;
      while (s < endPtr) {
	printRangeChar(buff, *s);
	s++;
      }
    } else if (op == ORSTART || op == LSTAR || op == STAR ||
	       op == BRANCH || op == LBRANCH) {
      int idx = UCHAR(*s);
      if (idx != UNOPTIMIZED) {
	if (idx < CS_LAST) {
	  sprintf(buff+strlen(buff), " %s", bistr[idx]);
	} else {
	  idx -= CS_LAST;
	  sprintf(buff+strlen(buff), "#%d", idx);
	  dumpRange(buff, ANYOFPTR(r->regclasses, idx));
	}
      }
      s++;
    } else if (op == BIANYOF) {
      sprintf(buff+strlen(buff), " %s", bistr[*s]);
      s++;
    } else if (op == ANYOF) {
      sprintf(buff+strlen(buff), "#%d", *s);
      dumpRange(buff, ANYOFPTR(r->regclasses, *s));
      s++;
    } else if (op == ISNOTCHAR) {
      strcat(buff, "[^");
      printRangeChar(buff, UCHAR(*s));
      strcat(buff, "]");
      s++;
    } else if (op == OPEN || op == CLOSE || op == BACKREF) {
      sprintf(buff+strlen(buff), "#%d", (int)UCHAR(*s));
      s++;
    } else if (op == REPEAT  || op == LREPEAT) {
      int idx = UCHAR(s[2]);
      sprintf(buff+strlen(buff), "%d,%d",
	      (int)UCHAR(s[0]),
	      (int)UCHAR(s[1]));
      if (idx != UNOPTIMIZED) {
	if (idx < CS_LAST) {
	  sprintf(buff+strlen(buff), " %s", bistr[idx]);
	} else {
	  idx -= CS_LAST;
	  sprintf(buff+strlen(buff), " #%d", idx);
	  dumpRange(buff, ANYOFPTR(r->regclasses, idx));
	}
      }
      s += 3;
    }
    if (Tcl_ListObjAppendElement(interp, resObjPtr, Tcl_NewStringObj(buff, -1))
	== TCL_ERROR) {
      Tcl_DecrRefCount(resObjPtr);
      return TCL_ERROR;
    }
  }

  /* Header fields of interest. */
  if (r->regstartnode) {
    sprintf(buff, "startnode: %s", regprop(r, r->regstartnode));
    if (Tcl_ListObjAppendElement(interp, resObjPtr, Tcl_NewStringObj(buff, -1))
	== TCL_ERROR) {
      Tcl_DecrRefCount(resObjPtr);
      return TCL_ERROR;
    }
  }
  if (r->regmust != NULL) {
    int i;
    sprintf(buff, "must have \"");
    for (i=0; i < r->regmlen; i++) {
      printRangeChar(buff, r->regmust[i]);
    }
    printRangeChar(buff, '"');
    if (Tcl_ListObjAppendElement(interp, resObjPtr, Tcl_NewStringObj(buff, -1))
	== TCL_ERROR) {
      Tcl_DecrRefCount(resObjPtr);
      return TCL_ERROR;
    }
  }
	  
  sprintf(buff, "flags: ");
  if (r->regflags & REGEXP_NOCASE) printRangeChar(buff, 'i');
  if (r->regflags & REGEXP_PERLMODE) printRangeChar(buff, 'p');
  if (r->regflags & REGEXP_MLINEMODE) printRangeChar(buff, 'm');
  if (r->regflags & REGEXP_QUICKCOMPILE) printRangeChar(buff, 'q');
  if (r->regflags & REGEXP_XMODE) printRangeChar(buff, 'x');
  if (Tcl_ListObjAppendElement(interp, resObjPtr, Tcl_NewStringObj(buff, -1))
      == TCL_ERROR) {
    Tcl_DecrRefCount(resObjPtr);
    return TCL_ERROR;
  }
  Tcl_SetObjResult(interp, resObjPtr);
  return TCL_OK;
}

/*
 - regprop - printable representation of opcode
 */
static char *
regprop(r, op)
regexp *r;
char *op;
{
	register char *p;
	static char buf[50];

        sprintf(buf, "%2d:", op - r->program);	/* Where. */

	switch (OP(op)) {
	case BACK:
		p = "BACK";
		break;
	case BOL:
		p = "BOL";
		break;
	case EOL:
		p = "EOL";
		break;
	case BOS:
		p = "BOS";
		break;
	case BOINPUT:
		p = "BOINPUT";
		break;
	case EOS:
		p = "EOS";
		break;
	case SOW:
		p = "SOW";
		break;
	case EOW:
		p = "EOW";
		break;
	case WBOUND:
		p = "WBOUND";
		break;
	case NWBOUND:
		p = "NWBOUND";
		break;
	case ANY:
		p = "ANY";
		break;
	case ANYOF:
		p = "ANYOF";
		break;
	case BIANYOF:
		p = "BIANYOF";
		break;
	case ORBRANCH:
		p = "ORBRANCH";
		break;
	case OREND:
		p = "OREND";
		break;
	case ORSTART:
		p = "ORSTART";
		break;
	case EXACTLY:
		p = "EXACTLY";
		break;
	case NCEXACTLY:
		p = "NCEXACTLY";
		break;
	case NOTHING:
		p = "NOTHING";
		break;
	case END:
		p = "END";
		break;
	case LOOK:
		p = "LOOK";
		break;
	case NLOOK:
		p = "NLOOK";
		break;
	case FASTLOOK:
		p = "FASTLOOK";
		break;
	case FASTNLOOK:
		p = "FASTNLOOK";
		break;
	case ENDLOOK:
		p = "ENDLOOK";
		break;
	case STAR:
		p = "STAR";
		break;
	case PLUS:
		p = "PLUS";
		break;
	case LSTAR:
		p = "LSTAR";
		break;
	case LPLUS:
		p = "LPLUS";
		break;
	case REPEAT:
		p = "REPEAT";
		break;
	case LREPEAT:
		p = "LREPEAT";
		break;
	case BRANCH:
		p = "BRANCH";
		break;
	case LBRANCH:
		p = "LBRANCH";
		break;
	case OPEN:
		p = "OPEN";
		break;
	case CLOSE:
		p = "CLOSE";
		break;
	case BACKREF:
		p = "BACKREF";
		break;
	case ISNOTCHAR:
		p = "ISNOTCHAR";
		break;
	default:
	  p = "corrupted opcode";
	  TclNewRegError("corrupted opcode");
	  break;
	}
	if (p != NULL)
		(void) strcat(buf, p);
	return(buf);
}

/*
 *----------------------------------------------------------------------
 *
 * TclNewRegError --
 *
 *	This procedure is invoked by the regexp code when an error
 *	occurs.  It saves the error message so it can be seen by the
 *	code that called Spencer's code.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The value of "string" is saved in "errMsg".
 *
 *----------------------------------------------------------------------
 */

void
TclNewRegError(string)
    char *string;			/* Error message. */
{
    errMsg = string;
}

char *
TclNewGetRegError()
{
    return errMsg;
}
